# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# duplicity is free software; you can redistribute it and/or modify
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
A library for interfacing with the librsync library to generate and apply deltas,
calculate signatures, and construct file-like operations.

This module provides a Python interface for the librsync library, enabling users
to calculate signatures, generate deltas, and incrementally apply deltas to files.
It uses both a Python implementation and a compiled C implementation depending on
availability. The core operations are modelled through file-like objects for
incremental processing.

Classes:

- librsyncError: Custom exception class for librsync-related errors.
- LikeFile: Abstract file-like object used as a base for signature, delta, and
  patched files.
- DeltaFile: File-like object to generate deltas based on a signature and a new file.
- PatchedFile: File-like object to apply deltas incrementally to create a patched file.
- SigGenerator: Class used to calculate signatures with an interface similar to
  the md5 module.

Functions:

- get_version: Retrieve the version of the underlying librsync library.

Attributes:

- blocksize: Integer denoting the block size used in librsync operations.
"""

import array
import os

from duplicity import log


def get_version():
    """
    Return the version of librsync
    """
    return _librsync.get_version()


# fmt:off
try:
    from . import _librsync
    log.Info(f"Using compiled-C version of _librsync {get_version()}")
except (ImportError, Exception):
    from . import _librsync_ctypes as _librsync
    log.Info(f"Using pure-Python version of _librsync {get_version()}")
# fmt:on

if os.environ.get("READTHEDOCS") == "True":
    import unittest.mock as mock
    import duplicity

    duplicity._librsync = mock.MagicMock()

blocksize = _librsync.RS_JOB_BLOCKSIZE


class librsyncError(Exception):
    """Signifies error in internal librsync processing (bad signature, etc.)

    underlying _librsync.librsyncError's are regenerated using this
    class because the C-created exceptions are by default
    unPickleable.  There is probably a way to fix this in _librsync,
    but this scheme was easier.

    """

    pass


class LikeFile(object):
    """
    File-like object used by SigFile, DeltaFile, and PatchFile
    """

    mode = "rb"

    # This will be replaced in subclasses by an object with
    # appropriate cycle() method
    maker = None

    def __init__(self, infile, need_seek=None):
        """
        LikeFile initializer - zero buffers, set eofs off
        """
        self.check_file(infile, need_seek)
        self.infile = infile
        self.closed = self.infile_closed = None
        self.inbuf = b""
        self.outbuf = array.array("b")
        self.eof = self.infile_eof = None

    def check_file(self, file, need_seek=None):
        """
        Raise type error if file doesn't have necessary attributes
        """
        if not hasattr(file, "read"):
            raise TypeError("Basis file must have a read() method")
        if not hasattr(file, "close"):
            raise TypeError("Basis file must have a close() method")
        if need_seek and not hasattr(file, "seek"):
            raise TypeError("Basis file must have a seek() method")

    def read(self, length=-1):
        """
        Build up self.outbuf, return first length bytes
        """
        if length == -1:
            while not self.eof:
                self._add_to_outbuf_once()
            real_len = len(self.outbuf)
        else:
            while not self.eof and len(self.outbuf) < length:
                self._add_to_outbuf_once()
            real_len = min(length, len(self.outbuf))

        return_val = self.outbuf[:real_len].tobytes()
        del self.outbuf[:real_len]
        return return_val

    def _add_to_outbuf_once(self):
        """
        Add one cycle's worth of output to self.outbuf
        """
        if not self.infile_eof:
            self._add_to_inbuf()
        try:
            self.eof, len_inbuf_read, cycle_out = self.maker.cycle(self.inbuf)
        except _librsync.librsyncError as e:
            raise librsyncError(str(e))
        self.inbuf = self.inbuf[len_inbuf_read:]
        self.outbuf.frombytes(cycle_out)

    def _add_to_inbuf(self):
        """
        Make sure len(self.inbuf) >= blocksize
        """
        assert not self.infile_eof, "infile_eof already set"
        while len(self.inbuf) < blocksize:
            new_in = self.infile.read(blocksize)
            if not new_in:
                self.infile_eof = 1
                assert not self.infile.close(), "self.infile failed to close"
                self.infile_closed = 1
                break
            self.inbuf += new_in

    def close(self):
        """
        Close infile
        """
        if not self.infile_closed:
            assert not self.infile.close(), "self.infile failed to close"
        self.closed = 1


class DeltaFile(LikeFile):
    """
    File-like object which incrementally generates a librsync delta
    """

    def __init__(self, signature, new_file):
        """
        Initialize a DeltaFile.

        :param signature: The signature to use, either as bytes or a file-like object.
        :param new_file: The new file to generate a delta for (file-like object).
        :raises librsyncError: If the signature is invalid.
        """
        LikeFile.__init__(self, new_file)
        if isinstance(signature, bytes):
            sig_string = signature
        else:
            self.check_file(signature)
            sig_string = signature.read()
            assert not signature.close(), "signature failed to close"
        try:
            self.maker = _librsync.new_deltamaker(sig_string)
        except _librsync.librsyncError as e:
            raise librsyncError(str(e))


class PatchedFile(LikeFile):
    """
    File-like object which applies a librsync delta incrementally
    """

    def __init__(self, basis_file, delta_file):
        """
        Initialize a PatchedFile.

        :param basis_file: The original file to patch. Must be a true Python
            file or have a ``fileno()`` method (except when using the ctypes backend).
        :param delta_file: The delta file to apply (file-like object).
        :raises TypeError: If ``basis_file`` is not a valid file object.
        :raises librsyncError: If an error occurs during patching.
        """
        LikeFile.__init__(self, delta_file)
        if hasattr(_librsync, "SigMaker"):
            # ctypes backend doesn't need fileno()
            self.maker = _librsync.new_patchmaker(basis_file)
        else:
            try:
                basis_file.fileno()
            except Exception as e:
                """tempfile.TemporaryFile() only guarantees a true file
                object on posix platforms. on cygwin/windows a file-like
                object whose file attribute is the underlying true file
                object is returned.
                """
                if hasattr(basis_file, "file") and hasattr(basis_file.file, "fileno"):
                    basis_file = basis_file.file
                else:
                    raise TypeError(
                        _(
                            "basis_file must be a (true) file or an object whose "
                            "file attribute is the underlying true file object"
                        )
                    )
            try:
                self.maker = _librsync.new_patchmaker(basis_file)
            except _librsync.librsyncError as e:
                raise librsyncError(str(e))


class SigGenerator(object):
    """Calculate signature.

    Interface is like md5 module

    """

    def __init__(self, blocksize=_librsync.RS_DEFAULT_BLOCK_LEN):
        """
        Return new signature instance
        """
        try:
            self.sig_maker = _librsync.new_sigmaker(blocksize)
        except _librsync.librsyncError as e:
            raise librsyncError(str(e))
        self.gotsig = None
        self.buffer = b""
        self.sigstring_list = []

    def update(self, buf):
        """
        Add data to be included in the signature calculation.

        :param bytes buf: The data to add.
        :raises librsyncError: If the signature has already been generated or
            if an internal error occurs.
        """
        if self.gotsig:
            raise librsyncError("SigGenerator already provided signature")
        self.buffer += buf
        while len(self.buffer) >= blocksize:
            if self.process_buffer():
                raise librsyncError("Premature EOF received from sig_maker")

    def process_buffer(self):
        """
        Run self.buffer through sig_maker, add to self.sig_string
        """
        try:
            eof, len_buf_read, cycle_out = self.sig_maker.cycle(self.buffer)
        except _librsync.librsyncError as e:
            raise librsyncError(str(e))
        self.buffer = self.buffer[len_buf_read:]
        self.sigstring_list.append(cycle_out)
        return eof

    def getsig(self):
        """
        Return the calculated signature over all added data.

        :returns: The signature as bytes.
        :rtype: bytes
        """
        while not self.process_buffer():
            pass  # keep running until eof
        return b"".join(self.sigstring_list)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
# Copyright 2014 Aaron Whitehouse
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Iterate exactly the requested files in a directory.

Parses includes and excludes to yield correct files.  More
documentation on what this code does can be found on the man page.
"""

import os
import re
import stat
import sys
from textwrap import dedent

from duplicity import config
from duplicity import diffdir
from duplicity import errors
from duplicity import log
from duplicity import util
from duplicity.globmatch import FilePrefixError
from duplicity.globmatch import GlobbingError
from duplicity.globmatch import select_fn_from_glob
from duplicity.path import Path


class Select(object):
    """Iterate appropriate Paths in given directory

    This class acts as an iterator on account of its next() method.
    Basically, it just goes through all the files in a directory in
    order (depth-first) and subjects each file to a bunch of tests
    (selection functions) in order.  The first test that includes or
    excludes the file means that the file gets included (iterated) or
    excluded.  The default is include, so with no tests we would just
    iterate all the files in the directory in order.

    The one complication to this is that sometimes we don't know
    whether or not to include a directory until we examine its
    contents.  For instance, if we want to include all the ``**.py``
    files.  If /home/ben/foo.py exists, we should also include /home
    and /home/ben, but if these directories contain no ``**.py`` files,
    they shouldn't be included.  For this reason, a test may not
    include or exclude a directory, but merely "scan" it.  If later a
    file in the directory gets included, so does the directory.

    As mentioned above, each test takes the form of a selection
    function.  The selection function takes a path, and returns:

    None - means the test has nothing to say about the related file
    0 - the file is excluded by the test
    1 - the file is included
    2 - the test says the file (must be directory) should be scanned

    Also, a selection function f has a variable f.exclude which should
    be true if f could potentially exclude some file. This is used
    to signal an error if the last function only includes, which would
    be redundant and presumably isn't what the user intends.
    """

    def __init__(self, path):
        """
        Initializer, called with Path of root directory
        """
        assert isinstance(path, Path), f"'path' must be a Path instance, got {type(path).__name__}: {path}"
        self.selection_functions = []
        self.rootpath = path
        self.prefix = self.rootpath.uc_name
        self.files_from = None

    def __iter__(self):  # pylint: disable=non-iterator-returned
        return self

    def __next__(self):
        return next(self.iter)

    def set_iter(self):
        """
        Initialize generator, prepare to iterate.
        """
        # Externally-accessed method
        self.rootpath.setdata()  # this may have changed since Select init
        self.iter = self.Iterate(self.rootpath)
        return self

    def Iterate(self, path):
        """Return iterator yielding paths in path

        This function looks a bit more complicated than it needs to be
        because it avoids extra recursion (and no extra function calls
        for non-directory files) while still doing the "directory
        scanning" bit.

        """

        # Only called by set_iter. Internal.
        def error_handler(exc, path, filename):  # pylint: disable=unused-argument
            fullpath = os.path.join(path.name, filename)
            try:
                mode = os.stat(fullpath)[stat.ST_MODE]
                if stat.S_ISSOCK(mode):
                    log.Info(
                        _("Skipping socket %s") % os.fsdecode(fullpath),
                        log.InfoCode.skipping_socket,
                        util.escape(fullpath),
                    )
                else:
                    log.Warn(
                        _("Error initializing file %s") % os.fsdecode(fullpath),
                        log.WarningCode.cannot_iterate,
                        util.escape(fullpath),
                    )
            except OSError:
                log.Warn(
                    _("Error accessing possibly locked file %s") % os.fsdecode(fullpath),
                    log.WarningCode.cannot_stat,
                    util.escape(fullpath),
                )
            return None

        def dir_scanner(path):
            """Generator of files to be included under from filesystems

            Yields (subpath, scan) where scan indicates that path is a
            directory and should be included *if* something inside is
            subsequently included (i.e. generated later)

            This generator applies the configured selection functions and
            so will not yield paths which are to be excluded.

            If --files-from has been specified, this will yield paths from
            that filelist only, subject to any selection functions.

            """
            # Only called by Iterate. Internal.
            from duplicity import robust  # TODO: avoid circ. dep. issue

            if self.files_from is None:
                files = robust.listpath(path)
            elif path.uc_name in self.files_from:
                files = self.files_from[path.uc_name]
            else:
                return

            for filename in files:
                new_path = robust.check_common_error(error_handler, Path.append, (path, filename))
                if new_path:
                    s = self.Select(new_path)
                    if (new_path.type in ["reg", "dir"] and not os.access(new_path.name, os.R_OK)) and (
                        s == 1 or s == 2
                    ):
                        # Path is a file or folder that cannot be read, but
                        # should be included or scanned.
                        log.Warn(
                            _("Error accessing possibly locked file %s") % new_path.uc_name,
                            log.WarningCode.cannot_read,
                            util.escape(new_path.name),
                        )
                        if diffdir.stats:
                            diffdir.stats.Errors += 1
                    elif s == 1:
                        # Should be included
                        yield new_path, False
                    elif s == 2 and new_path.isdir():
                        # Is a directory that should be scanned
                        yield new_path, True

        if not path.type:
            # base doesn't exist
            log.Warn(_("WARNING: base %s doesn't exist, continuing") % path.uc_name)
            return
        log.Debug(_("Selecting %s") % path.uc_name)
        yield path
        if not path.isdir():
            return

        dirs_to_scan = [dir_scanner(path)]
        dirs_deferred = []
        while dirs_to_scan:
            try:
                subpath, scan = next(dirs_to_scan[-1])
            except StopIteration:
                dirs_to_scan.pop()
                if dirs_deferred:
                    dirs_deferred.pop()
                continue
            if scan:
                dirs_deferred.append(subpath)
                dirs_to_scan.append(dir_scanner(subpath))
            else:
                if dirs_deferred:
                    for delayed_path in dirs_deferred:
                        log.Log(_("Selecting %s") % delayed_path.uc_name, 6)
                        yield delayed_path
                    del dirs_deferred[:]
                log.Debug(_("Selecting %s") % subpath.uc_name)
                yield subpath
                if subpath.isdir():
                    dirs_to_scan.append(dir_scanner(subpath))

    def Select(self, path):
        """
        Run through the selection functions and return dominant val 0/1/2
        """
        # Only used by diryield and tests. Internal.
        log.Debug(f"Selection: examining path {path.uc_name}")
        if not self.selection_functions:
            log.Debug("Selection:     + no selection functions found. Including")
            return 1
        scan_pending = False
        for sf in self.selection_functions:
            result = sf(path)
            log.Debug(f"Selection:     result: {str(result):4} from function: {sf.name}")
            if result == 2:
                # Selection function says that the path should be scanned for matching files, but keep going
                # through the selection functions looking for a real match (0 or 1).
                scan_pending = True
            elif result == 0 or result == 1:
                # A real match found, no need to try other functions.
                break

        if scan_pending and result != 1:
            # A selection function returned 2 and either no real match was
            # found or the highest-priority match was 0
            result = 2
        if result is None:
            result = 1

        if result == 0:
            log.Debug("Selection:     - excluding file")
        elif result == 1:
            log.Debug("Selection:     + including file")
        else:
            assert result == 2, "Selection result must be 2 (scan) when not including or excluding"
            log.Debug("Selection:     ? scanning directory for matches")

        return result

    def ParseArgs(self, argtuples, filelists):
        """
        Create selection functions based on a list of tuples.

        The tuples are created when the initial commandline arguments
        are read. They have the form (option string, additional
        argument) except for the filelist tuples, which should be
        (option-string, (additional argument, filelist_fp)).

        :param list argtuples: List of (option, argument) tuples.
        :param list filelists: List of file objects for filelists.
        :raises GlobbingError: If a globbing pattern is invalid.
        """
        # Sanity checks on --filter-* options for the benefit of users
        if argtuples and argtuples[-1][0].startswith("--filter-"):
            raise errors.FatalError(
                dedent(_("""\
                The last file selection option is the filter option %s, which will have no
                effect as there are no subsequent file selection options. Exiting because
                this probably isn't what you meant.""")) % (argtuples[-1][0],),
                log.ErrorCode.trailing_filter,
            )
        f_opt = set(opt[0] for opt in argtuples if opt[0].startswith("--filter-"))
        f_def = ("--filter-globbing", "--filter-strictcase")
        if f_opt and all(opt in f_def for opt in f_opt):
            raise errors.FatalError(
                dedent(_("""\
                Only these filter mode options were specified:
                    %s
                Case sensitive globbing is the default behaviour and so this has no effect.
                Exiting because this probably isn't what you meant.""")) % (", ".join(f_opt),),
                log.ErrorCode.redundant_filter,
            )

        # Called by cli_main.py set_selection. External.
        filelists_index = 0
        mode = "globbing"
        no_case = False
        try:
            for opt, arg in argtuples:
                if opt == "--exclude":
                    self.add_selection_func(self.general_get_sf(arg, 0, mode, ignore_case=no_case))
                elif opt == "--exclude-if-present":
                    self.add_selection_func(self.present_get_sf(arg, 0), add_to_start=True)
                elif opt == "--exclude-device-files":
                    self.add_selection_func(self.devfiles_get_sf(), add_to_start=True)
                elif opt == "--exclude-filelist":
                    for sf in self.filelist_general_get_sfs(filelists[filelists_index], 0, arg, mode, no_case):
                        self.add_selection_func(sf)
                    filelists_index += 1
                elif opt == "--exclude-other-filesystems":
                    self.add_selection_func(self.other_filesystems_get_sf(0))
                elif opt == "--exclude-regexp":
                    self.add_selection_func(self.regexp_get_sf(arg, 0, no_case))
                elif opt == "--exclude-older-than":
                    self.add_selection_func(self.exclude_older_get_sf(arg))
                elif opt == "--filter-strictcase":
                    no_case = False
                elif opt == "--filter-globbing":
                    mode = "globbing"
                elif opt == "--filter-ignorecase":
                    no_case = True
                elif opt == "--filter-literal":
                    mode = "literal"
                elif opt == "--filter-regexp":
                    mode = "regex"
                elif opt == "--files-from":
                    self.parse_files_from(filelists[filelists_index], arg)
                    filelists_index += 1
                elif opt == "--include":
                    self.add_selection_func(self.general_get_sf(arg, 1, mode, no_case))
                elif opt == "--include-filelist":
                    for sf in self.filelist_general_get_sfs(filelists[filelists_index], 1, arg, mode, no_case):
                        self.add_selection_func(sf)
                    filelists_index += 1
                elif opt == "--include-regexp":
                    self.add_selection_func(self.regexp_get_sf(arg, 1, no_case))
                else:
                    assert 0, f"Bad selection option {opt}"
        except GlobbingError as e:
            self.parse_catch_error(e)
        assert filelists_index == len(filelists), "Number of filelists consumed does not match provided list length"
        self.parse_last_excludes()

    def parse_catch_error(self, exc):
        """
        Deal with selection error exc
        """
        # Internal, used by ParseArgs.
        if isinstance(exc, FilePrefixError):
            raise errors.FatalError(
                dedent(_("""\
                Fatal Error: The file specification
                    %s
                cannot match any files in the base directory
                    %s
                Useful file specifications begin with the base directory or some
                pattern (such as '**') which matches the base directory.""")) % (exc, self.prefix),
                log.ErrorCode.file_prefix_error,
            )
        elif isinstance(exc, GlobbingError):
            raise errors.FatalError(
                _("Fatal Error while processing expression\n" "%s") % exc,
                log.ErrorCode.globbing_error,
            )
        else:
            raise  # pylint: disable=misplaced-bare-raise

    def parse_files_from(self, filelist_fp, list_name):
        """Loads an explicit list of files to backup from a filelist, building
        a dictionary of directories and their contents which can be used later
        to emulate a filesystem walk over the listed files only.

        Each specified path is unwound to identify the parents folder(s) as these
        are implicitly to be included.

        Paths read are not to be stripped, checked for comments, etc. Every
        character on each line is significant and treated as part of the path.
        """
        # Internal. Used by ParseArgs.
        log.Notice(_("Reading files to backup from list %s") % list_name)
        separator = config.null_separator and "\0" or "\n"
        filelist = {}
        absolute_path = None
        is_root = self.rootpath.get_parent_dir() == self.rootpath
        for line in filelist_fp.read().split(separator):
            if not line:  # skip blanks
                continue
            if line.startswith("/"):  # no absolute paths thanks
                absolute_path = line
                break
            while line:
                dirname, basename = os.path.split(line)
                path = os.path.join(self.rootpath.uc_name, dirname)
                # strip the trailing separator unless in root
                if not is_root or len(dirname) > 0:
                    path = path.rstrip(os.path.sep)
                if path not in filelist:
                    filelist[path] = set()
                if isinstance(basename, str):
                    filelist[path].add(os.fsencode(basename))
                else:
                    filelist[path].add(basename)
                line = dirname

        if absolute_path:
            raise errors.FatalError(
                dedent(_("""\
                Files-from list contains the absolute path:
                    %s
                All paths specified in a files-from list must be given relative to the backup
                source path.""")) % (absolute_path,),
                log.ErrorCode.absolute_files_from,
            )

        if not filelist:
            raise errors.FatalError(
                dedent(_("""\
                Files-from list specified which contains no files, the backup will be empty as
                a result. Exiting as this probably isn't what you meant,""")),
                log.ErrorCode.empty_files_from,
            )

        self.files_from = {d: sorted(f) for d, f in filelist.items()}

    def parse_last_excludes(self):
        """
        Exit with error if last selection function isn't an exclude
        """
        # Internal. Used by ParseArgs.
        if self.selection_functions and not self.selection_functions[-1].exclude:
            raise errors.FatalError(
                dedent(_("""\
                Last selection expression:
                    %s
                only specifies that files be included.  Because the default is to
                include all files, the expression is redundant.  Exiting because this
                probably isn't what you meant.""")) % (self.selection_functions[-1].name,),
                log.ErrorCode.redundant_inclusion,
            )

    def add_selection_func(self, sel_func, add_to_start=None):
        """
        Add another selection function at the end or beginning.

        :param callable sel_func: The selection function to add.
        :param bool add_to_start: If True, add to the beginning of the list.
        """
        # Internal. Used by ParseArgs.
        if add_to_start:
            self.selection_functions.insert(0, sel_func)
        else:
            self.selection_functions.append(sel_func)

    def filelist_sanitise_line(self, line, include_default):
        """
        Sanitises lines of both normal and globbing filelists, returning
        (line, include) and line=None if blank/comment

        The aim is to parse filelists in a consistent way, prior to the
        interpretation of globbing statements. The function removes
        whitespace, comment lines and processes modifiers (leading +/-)
        and quotes.
        """
        # Internal. Used by filelist_general_get_sfs

        line = line.strip()
        if not line:  # skip blanks
            return None, include_default
        if line.startswith("#"):  # skip full-line comments
            return None, include_default

        include = include_default
        if line.startswith("+ "):
            # Check for "+ " or "- " syntax
            include = 1
            line = line[2:]
        elif line.startswith("- "):
            include = 0
            line = line[2:]

        if (line.startswith("'") and line.endswith("'")) or (line.startswith('"') and line.endswith('"')):
            line = line[1:-1]

        return line, include

    def filelist_general_get_sfs(self, filelist_fp, inc_default, list_name, mode="globbing", ignore_case=False):
        """Return list of selection functions by reading fileobj

        filelist_fp should be an open file object
        inc_default is true if this is an include list
        list_name is just the name of the list, used for logging
        mode indicates whether to glob, regex, or not

        """
        # Internal. Used by ParseArgs.
        log.Notice(_("Reading %s filelist %s") % (mode, list_name))
        separator = config.null_separator and "\0" or "\n"
        try:
            filelist_fp.seek(0)
        except Exception as e:
            pass
        for line in filelist_fp.read().split(separator):
            line, include = self.filelist_sanitise_line(line, inc_default)
            if not line:
                # Skip blanks and comment lines
                continue
            yield self.general_get_sf(line, include, mode, ignore_case)

    def other_filesystems_get_sf(self, include):
        """
        Return selection function matching files on other filesystems
        """
        # Internal. Used by ParseArgs and unit tests.
        assert include == 0 or include == 1, "'include' must be 0 (exclude) or 1 (include)"
        root_devloc = self.rootpath.getdevloc()

        def sel_func(path):
            if path.exists() and path.getdevloc() != root_devloc:
                return include
            else:
                return None

        sel_func.exclude = not include
        sel_func.name = "Match other filesystems"
        return sel_func

    def regexp_get_sf(self, regexp_string, include, ignore_case=False):
        """
        Return selection function given by regexp_string
        """
        # Internal. Used by ParseArgs and unit tests.
        assert include == 0 or include == 1, "'include' must be 0 (exclude) or 1 (include)"

        flags = 0
        if ignore_case:
            flags = re.IGNORECASE

        try:
            regexp = re.compile(regexp_string, flags)
        except Exception:
            log.Warn(_("Error compiling regular expression %s") % regexp_string)
            raise

        def sel_func(path):
            if regexp.search(path.uc_name):
                return include
            else:
                return None

        sel_func.exclude = not include
        sel_func.name = (
            f"regular expression {include and 'include' or 'exclude'} "
            f"{ignore_case and 'no-' or ''}case: {regexp_string}"
        )
        return sel_func

    def devfiles_get_sf(self):
        """
        Return a selection function to exclude all dev files
        """

        # Internal. Used by ParseArgs.
        def sel_func(path):
            if path.isdev():
                return 0
            else:
                return None

        sel_func.exclude = 1
        sel_func.name = "Exclude device files"
        return sel_func

    def general_get_sf(self, pattern_str, include, mode="globbing", ignore_case=False):
        """Return selection function given by a pattern string

        The selection patterns are interpretted in accordance with the mode
        argument, "globbing", "literal", or "regex".

        The 'ignorecase:' prefix is a legacy feature which historically lived on
        the globbing code path and was only ever documented as working for globs.
        """

        # Internal. Used by ParseArgs, filelist_general_get_sfs and unit tests.
        assert include == 0 or include == 1, "'include' must be 0 (exclude) or 1 (include)"
        assert isinstance(pattern_str, str), "pattern_str must be a string"

        # legacy prefix applies *only* in globbing mode
        if mode == "globbing" and pattern_str.lower().startswith("ignorecase:"):
            pattern_str = pattern_str[len("ignorecase:") :]
            pattern_str = os.fsdecode(pattern_str).casefold()
            ignore_case = True

        if mode == "globbing":
            return self.glob_get_sf(pattern_str, include, ignore_case)
        elif mode == "literal":
            return self.literal_get_sf(pattern_str, include, ignore_case)
        elif mode == "regex":
            return self.regexp_get_sf(pattern_str, include, ignore_case)
        else:
            assert 0, f"Bad selection mode {mode}"

    def present_get_sf(self, filename, include):
        """
        Return selection function given by existence of a file in a directory
        """
        # Internal. Used by ParseArgs.
        assert include == 0 or include == 1, "'include' must be 0 (exclude) or 1 (include)"

        from duplicity.robust import check_common_error  # TODO: avoid circ. dep. issue

        def exclude_sel_func(path):
            # do not follow symbolic links when checking for file existence!
            if path.isdir():

                def error_handler(_exc, _filename):
                    # Path is not read accessible
                    # TODO: Ideally this error would only show if the folder
                    # was ultimately included by the full set of selection
                    # functions. Currently this will give an error for any
                    # locked directory within the folder being backed up.
                    log.Warn(
                        _("Error accessing possibly locked file %s") % path.uc_name,
                        log.WarningCode.cannot_read,
                        util.escape(path.uc_name),
                    )
                    if diffdir.stats:
                        diffdir.stats.Errors += 1
                    return False

                if check_common_error(error_handler, path.contains, [filename]):
                    return 0
                else:
                    return None

        if include == 0:
            sel_func = exclude_sel_func
        else:
            raise errors.FatalError(
                "--include-if-present not implemented (would it make sense?).",
                log.ErrorCode.not_implemented,
            )

        sel_func.exclude = not include
        sel_func.name = f"Command-line {include and 'include-if-present' or 'exclude-if-present'} filename: {filename}"
        return sel_func

    def glob_get_sf(self, glob_str, include, ignore_case=False):
        """
        Return selection function based on glob_str
        """
        assert isinstance(
            glob_str, str
        ), f"The glob string {glob_str.decode(sys.getfilesystemencoding(), 'ignore')} is not unicode"

        # Check to make sure prefix is ok, i.e. the glob string is within
        # the root folder being backed up
        if not select_fn_from_glob(glob_str, 1, ignore_case)(self.rootpath):
            # file_prefix_selection == 1 (include) or 2 (scan)
            raise FilePrefixError(glob_str)

        if glob_str == "**":
            sel_func = lambda path: include
        else:
            sel_func = select_fn_from_glob(glob_str, include, ignore_case)

        sel_func.exclude = not include
        sel_func.name = f"shell glob {include and 'include' or 'exclude'} {ignore_case and 'no-' or ''}case: {glob_str}"

        return sel_func

    def literal_get_sf(self, lit_str, include, ignore_case=False):
        """Return a selection function that matches a literal string while
        still including the contents of any folders which are matched
        """

        # Check to make sure prefix is ok, see also glob_get_sf()
        if not self.select_fn_from_literal(lit_str, 1, ignore_case)(self.rootpath):
            raise FilePrefixError(lit_str)

        sel_func = self.select_fn_from_literal(lit_str, include, ignore_case)
        sel_func.exclude = not include
        sel_func.name = (
            f"literal string {include and 'include' or 'exclude'} " f"{ignore_case and 'no-' or ''}case: {lit_str}"
        )
        return sel_func

    def exclude_older_get_sf(self, date):
        """
        Return selection function based on files older than modification date
        """

        # Internal. Used by ParseArgs.

        def sel_func(path):
            if not path.isreg():
                return None
            try:
                if os.path.getmtime(path.name) < date:
                    return 0
            except OSError as e:
                pass  # this is probably only on a race condition of file being deleted
            return None

        sel_func.exclude = True
        sel_func.name = f"Select older than {date}"
        return sel_func

    def select_fn_from_literal(self, lit_str, include, ignore_case=False):
        """Return a function test_fn(path) which test where a path matches a
        literal string. See also select_fn_from_blog() in globmatch.py

        This function is separated from literal_get_sf() so that it can be used
        to test the prefix without creating a loop.

        """
        # TODO: this doesn't need to be part of the Select class type, but not
        #       sure where else to put it?
        if lit_str != "/" and lit_str[-1] == "/":
            lit_str = lit_str[:-1]

        if ignore_case:
            lit_str = os.fsdecode(lit_str).casefold()

        def test_fn(path):
            # TODO: caller to do this once, rather than for every path?
            if ignore_case:
                uc_name = os.fsdecode(path.uc_name).casefold()
            else:
                uc_name = path.uc_name

            if uc_name == lit_str:
                return include
            elif uc_name.startswith(lit_str) and uc_name[len(lit_str)] == "/":
                return include
            elif lit_str.startswith(uc_name) and lit_str[len(uc_name)] == "/" and include == 1:
                return 2
            else:
                return None

        return test_fn


class _ArchivePath(object):
    """
    Adapter exposing an archive ROPath with an archive-relative name.
    """

    def __init__(self, ropath):
        self.ropath = ropath
        self.index = ropath.index
        if self.index:
            self.uc_name = os.fsdecode(b"/".join(self.index))
        else:
            self.uc_name = ""

    def __getattr__(self, attr):
        return getattr(self.ropath, attr)


class ArchiveSelect(Select):
    """Apply selection rules to archive-relative restore paths.

    Restore paths are matched in the namespace printed by list-current-files
    and accepted by --path-to-restore, not in the original source filesystem
    namespace used by backup selection.
    """

    def __init__(self):
        super().__init__(Path(b"."))

    @staticmethod
    def _normalize_archive_pattern(pattern):
        while pattern.startswith("./"):
            pattern = pattern[2:]
        if pattern.startswith("/"):
            raise FilePrefixError(pattern)
        if any(part in (".", "..") for part in pattern.split("/") if part):
            raise FilePrefixError(pattern)
        return pattern

    def parse_catch_error(self, exc):
        """
        Deal with restore archive-relative selection errors.
        """
        if isinstance(exc, FilePrefixError):
            raise errors.FatalError(
                dedent(_("""\
                Fatal Error: Restore file selection pattern
                    %s
                is invalid. Restore file selection patterns must be relative to the
                archive root, using paths as printed by list-current-files, and must
                not be absolute or contain '.' or '..' path elements.""")) % (exc,),
                log.ErrorCode.file_prefix_error,
            )
        else:
            super().parse_catch_error(exc)

    def glob_get_sf(self, glob_str, include, ignore_case=False):
        """
        Return archive-relative selection function based on glob_str.
        """
        glob_str = self._normalize_archive_pattern(glob_str)
        if glob_str == "**":
            sel_func = lambda path: include
        else:
            sel_func = select_fn_from_glob(glob_str, include, ignore_case)

        sel_func.exclude = not include
        sel_func.name = (
            f"archive shell glob {include and 'include' or 'exclude'} " f"{ignore_case and 'no-' or ''}case: {glob_str}"
        )
        return sel_func

    def literal_get_sf(self, lit_str, include, ignore_case=False):
        """
        Return archive-relative selection function based on literal string.
        """
        lit_str = self._normalize_archive_pattern(lit_str)
        sel_func = self.select_fn_from_literal(lit_str, include, ignore_case)
        sel_func.exclude = not include
        sel_func.name = (
            f"archive literal string {include and 'include' or 'exclude'} "
            f"{ignore_case and 'no-' or ''}case: {lit_str}"
        )
        return sel_func

    def exclude_older_get_sf(self, date):
        """
        Return archive-relative selection function based on archived mtime.
        """

        def sel_func(path):
            if not path.isreg():
                return None
            if path.getmtime() < date:
                return 0
            return None

        sel_func.exclude = True
        sel_func.name = f"Select older than {date}"
        return sel_func


def get_restore_selection(argtuples, filelists):
    """
    Return an archive-relative selector for restore filtering.
    """
    sel = ArchiveSelect()
    sel.ParseArgs(argtuples, filelists)
    return sel


def filter_restore_path_iter(path_iter, sel):
    """
    Yield selected restore ROPaths, including deferred parent directories.
    """
    deferred_dirs = []
    excluded_dirs = []
    yielded_dirs = set()

    def is_ancestor(parent_index, child_index):
        return parent_index == child_index[: len(parent_index)]

    def prune_deferred_dirs(index):
        while deferred_dirs and not is_ancestor(deferred_dirs[-1].index, index):
            deferred_dirs.pop()

    def prune_excluded_dirs(index):
        while excluded_dirs and not is_ancestor(excluded_dirs[-1], index):
            excluded_dirs.pop()

    def is_excluded_by_parent(index):
        return excluded_dirs and is_ancestor(excluded_dirs[-1], index)

    def yield_deferred_dirs():
        for deferred_dir in deferred_dirs:
            if deferred_dir.index not in yielded_dirs:
                yielded_dirs.add(deferred_dir.index)
                yield deferred_dir
        del deferred_dirs[:]

    def close_unselected(ropath):
        fileobj = getattr(ropath, "fileobj", None)
        if fileobj:
            try:
                fileobj.close()
            except Exception:
                pass

    for ropath in path_iter:
        prune_deferred_dirs(ropath.index)
        prune_excluded_dirs(ropath.index)
        if is_excluded_by_parent(ropath.index):
            close_unselected(ropath)
            continue

        if ropath.index == ():
            if ropath.index not in yielded_dirs:
                deferred_dirs.append(ropath)
            continue

        result = sel.Select(_ArchivePath(ropath))
        if result == 1:
            yield from yield_deferred_dirs()
            if ropath.isdir():
                yielded_dirs.add(ropath.index)
            yield ropath
        elif result == 2 and ropath.isdir() and ropath.index not in yielded_dirs:
            deferred_dirs.append(ropath)
        else:
            if result == 0 and ropath.isdir():
                excluded_dirs.append(ropath.index)
            close_unselected(ropath)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright (c) 2014 Gu1
# Licensed under the MIT license

"""
Hubic identity handling for Pyrax.
"""

import configparser
import os
import re
import time
import urllib.parse

import requests
from requests.compat import quote
from requests.compat import quote_plus

try:
    import pyrax
    from pyrax.base_identity import BaseIdentity
    from pyrax.base_identity import Service
    import pyrax.exceptions as exc
except ImportError as e:
    raise BackendException(f"""Hubic backend requires the pyrax library available from Rackspace.
Exception: {str(e)}""")

OAUTH_ENDPOINT = "https://api.hubic.com/oauth/"
API_ENDPOINT = "https://api.hubic.com/1.0/"
TOKENS_FILE = os.path.expanduser("~/.hubic_tokens")


class BearerTokenAuth(requests.auth.AuthBase):
    """
    Bearer token authentication for requests.
    """

    def __init__(self, token):
        """
        Initialize BearerTokenAuth.

        :param str token: The bearer token.
        """
        self.token = token

    def __call__(self, req):
        req.headers["Authorization"] = f"Bearer {self.token}"
        return req


class HubicIdentity(BaseIdentity):
    """
    Hubic-specific identity class for Pyrax.
    """

    def _get_auth_endpoint(self):
        return ""

    def set_credentials(
        self,
        email,
        password,
        client_id,
        client_secret,
        redirect_uri,
        authenticate=False,
    ):
        """
        Sets the username and password directly.
        """
        self._email = email
        self._password = password
        self._client_id = client_id
        self.tenant_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        if authenticate:
            self.authenticate()

    def _read_credential_file(self, cfg):
        """
        Parses the credential file with Rackspace-specific labels.
        """
        self._email = cfg.get("hubic", "email")
        self._password = cfg.get("hubic", "password")
        self._client_id = cfg.get("hubic", "client_id")
        self.tenant_id = self._client_id
        self._client_secret = cfg.get("hubic", "client_secret")
        self._redirect_uri = cfg.get("hubic", "redirect_uri")

    def _parse_error(self, resp):
        if "location" not in resp.headers:
            return None
        query = urllib.parse.urlsplit(resp.headers["location"]).query
        qs = dict(urllib.parse.parse_qsl(query))
        return {"error": qs["error"], "error_description": qs["error_description"]}

    def _get_access_token(self, code):
        r = requests.post(
            f"{OAUTH_ENDPOINT}token/",
            data={
                "code": code,
                "redirect_uri": self._redirect_uri,
                "grant_type": "authorization_code",
            },
            auth=(self._client_id, self._client_secret),
        )
        if r.status_code != 200:
            try:
                err = r.json()
                err["code"] = r.status_code
            except Exception as e:
                err = {}

            raise exc.AuthenticationFailed(
                f"Unable to get oauth access token, wrong client_id or client_secret ? " f"({str(err)})"
            )

        oauth_token = r.json()

        config = configparser.ConfigParser()
        config.read(TOKENS_FILE)

        if not config.has_section("hubic"):
            config.add_section("hubic")

        if oauth_token["access_token"] is not None:
            config.set("hubic", "access_token", oauth_token["access_token"])
            with open(TOKENS_FILE, "wb") as configfile:
                config.write(configfile)
        else:
            raise exc.AuthenticationFailed(
                f"Unable to get oauth access token, wrong client_id or client_secret ? ({str(err)})"
            )

        if oauth_token["refresh_token"] is not None:
            config.set("hubic", "refresh_token", oauth_token["refresh_token"])
            with open(TOKENS_FILE, "wb") as configfile:
                config.write(configfile)
        else:
            raise exc.AuthenticationFailed("Unable to get the refresh token.")

        # removing username and password from .hubic_tokens
        if config.has_option("hubic", "email"):
            config.remove_option("hubic", "email")
            with open(TOKENS_FILE, "wb") as configfile:
                config.write(configfile)
            print("username has been removed from the .hubic_tokens file sent to the CE.")
        if config.has_option("hubic", "password"):
            config.remove_option("hubic", "password")
            with open(TOKENS_FILE, "wb") as configfile:
                config.write(configfile)
            print("password has been removed from the .hubic_tokens file sent to the CE.")

        return oauth_token

    def _refresh_access_token(self):
        config = configparser.ConfigParser()
        config.read(TOKENS_FILE)
        refresh_token = config.get("hubic", "refresh_token")

        if refresh_token is None:
            raise exc.AuthenticationFailed("refresh_token is null. Not acquiered before ?")

        success = False
        max_retries = 20
        retries = 0
        sleep_time = 30
        max_sleep_time = 3600

        while retries < max_retries and not success:
            r = requests.post(
                f"{OAUTH_ENDPOINT}token/",
                data={
                    "refresh_token": refresh_token,
                    "grant_type": "refresh_token",
                },
                auth=(self._client_id, self._client_secret),
            )
            if r.status_code != 200:
                if r.status_code == 509:
                    print("status_code 509: attempt #", retries, " failed")
                    retries += 1
                    time.sleep(sleep_time)
                    sleep_time = sleep_time * 2
                    if sleep_time > max_sleep_time:
                        sleep_time = max_sleep_time
                else:
                    try:
                        err = r.json()
                        err["code"] = r.status_code
                    except Exception as e:
                        err = {}

                    raise exc.AuthenticationFailed(
                        f"Unable to get oauth access token, wrong client_id or client_secret ? ({str(err)})"
                    )
            else:
                success = True

        if not success:
            raise exc.AuthenticationFailed(
                "All the attempts failed to get the refresh token: " "status_code = 509: Bandwidth Limit Exceeded"
            )

        oauth_token = r.json()

        if oauth_token["access_token"] is not None:
            return oauth_token
        else:
            raise exc.AuthenticationFailed("Unable to get oauth access token from json")

    def authenticate(self):
        config = configparser.ConfigParser()
        config.read(TOKENS_FILE)

        if config.has_option("hubic", "refresh_token"):
            oauth_token = self._refresh_access_token()
        else:
            r = requests.get(
                OAUTH_ENDPOINT
                + f"auth/?client_id={quote(self._client_id)}&redirect_uri={quote_plus(self._redirect_uri)}"
                f"&scope=credentials.r,account.r&response_type=code&state={pyrax.utils.random_ascii()}",
                allow_redirects=False,
            )
            if r.status_code != 200:
                raise exc.AuthenticationFailed(f"Incorrect/unauthorized client_id ({str(self._parse_error(r))})")

            try:
                from lxml import html as lxml_html
            except ImportError:
                lxml_html = None

            if lxml_html:
                oauth = lxml_html.document_fromstring(r.content).xpath('//input[@name="oauth"]')
                oauth = oauth[0].value if oauth else None
            else:
                oauth = re.search(
                    r'<input\s+[^>]*name=[\'"]?oauth[\'"]?\s+[^>]*value=[\'"]?(\d+)[\'"]?>',
                    r.content,
                )
                oauth = oauth.group(1) if oauth else None

            if not oauth:
                raise exc.AuthenticationFailed("Unable to get oauth_id from authorization page")

            if self._email is None or self._password is None:
                raise exc.AuthenticationFailed(
                    "Cannot retrieve email and/or password. " "Please run expresslane-hubic-setup.sh"
                )

            r = requests.post(
                f"{OAUTH_ENDPOINT}auth/",
                data={
                    "action": "accepted",
                    "oauth": oauth,
                    "login": self._email,
                    "user_pwd": self._password,
                    "account": "r",
                    "credentials": "r",
                },
                allow_redirects=False,
            )

            try:
                query = urllib.parse.urlsplit(r.headers["location"]).query
                code = dict(urllib.parse.parse_qsl(query))["code"]
            except Exception as e:
                raise exc.AuthenticationFailed("Unable to authorize client_id, " "invalid login/password ?")

            oauth_token = self._get_access_token(code)

        if oauth_token["token_type"].lower() != "bearer":
            raise exc.AuthenticationFailed("Unsupported access token type")

        r = requests.get(
            f"{API_ENDPOINT}account/credentials",
            auth=BearerTokenAuth(oauth_token["access_token"]),
        )

        swift_token = r.json()
        self.authenticated = True
        self.token = swift_token["token"]
        self.expires = swift_token["expires"]
        self.services["object_store"] = Service(
            self,
            {
                "name": "HubiC",
                "type": "cloudfiles",
                "endpoints": [{"public_url": swift_token["endpoint"]}],
            },
        )
        self.username = self.password = None



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escotodo
# Copyright 2007 Kenneth Loafman
# Copyright 2022 Thomas Kramer
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Xorriso (ISO-9660) backend.
"""

import fcntl
import os
import re
import shlex
import subprocess
import time

import duplicity.backend
from duplicity import progress
from duplicity import util
from duplicity.errors import BackendException
from duplicity.errors import FatalBackendException
from duplicity.errors import InvalidBackendURL


class Xorriso:
    """
    Wrapper around a xorriso subprocess.
    """

    def __init__(self, device, xorriso_path="xorriso", xorriso_args=None):
        self.device = device

        # Xorriso process
        self.proc = None

        # Default arguments for xorriso.
        self.xorriso_args = [
            "-abort_on",
            "FAILURE",
            "-return_with",
            "SORRY",
            "0",
            "-osirrox",
            "on",  # Enable copying from ISO to disk.
            "-calm_drive",
            "off",  # Don't immediately turn off device. Increases access speed for next action.abs
            "-joliet",
            "on",
        ]

        if xorriso_args is not None:
            self.xorriso_args.extend(xorriso_args)

        self.__start_subprocess(
            [xorriso_path]
            + self.xorriso_args
            + [
                "-dev",
                self.device,
                "-dialog",
                "on",  # Enable interactive mode
            ]
        )

        stdout, stderr = self.__recv_stdout_stderr()
        self.__handle_xorriso_error(stderr)

        stdout, stderr = self.__send_cmd("-version")  # Test connectivity to subprocess.

    def __start_subprocess(self, commandline):
        def setNonBlocking(fd):
            flags = fcntl.fcntl(fd, fcntl.F_GETFL)
            flags = flags | os.O_NONBLOCK
            fcntl.fcntl(fd, fcntl.F_SETFL, flags)

        try:
            p = subprocess.Popen(
                commandline,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            self.proc = p
        except FileNotFoundError as err:
            raise FatalBackendException(f"Xorriso binary not found: {self.xorriso_cmd}")

        setNonBlocking(self.proc.stdout)
        setNonBlocking(self.proc.stderr)

    def __send_cmd(self, *args):
        # Drain stdout and stderr

        while self.proc.stdout.readline():
            pass

        while self.proc.stderr.readline():
            pass

        try:
            for arg in args:
                self.proc.stdin.write(shlex.quote(arg).encode("utf-8"))
                self.proc.stdin.write(b" ")
            self.proc.stdin.write(b"\n")
            self.proc.stdin.flush()
        except BrokenPipeError as e:
            raise FatalBackendException("BrokenPipe: lost connection to xorriso subprocess")

        stdout, stderr = self.__recv_stdout_stderr()

        self.__handle_xorriso_error(stderr)

        return stdout, stderr

    def __recv_stdout_stderr(self):
        stdout = []
        stderr = []
        while True:
            if self.proc.poll() is not None:
                # Process terminated
                break

            no_input = True

            while True:
                line = self.proc.stdout.readline().decode("utf-8").strip()
                if line:
                    stdout.append(line)
                else:
                    no_input = False
                    break

            while True:
                line = self.proc.stderr.readline().decode("utf-8").strip()
                if line:
                    stderr.append(line)
                else:
                    no_input = False
                    break

            if stderr and stderr[-1] == "enter option and arguments :":
                break

            if no_input:
                time.sleep(0.1)

        return stdout, stderr

    def __handle_xorriso_error(self, stderr):
        """
        Detect errors from stderr of xorriso.
        Convert the errors into exceptions.
        """

        lines = stderr

        stderr = "\n".join(lines)

        is_fatal = "FATAL" in stderr
        is_failure = "FAILURE" in stderr

        if is_fatal:
            msg = "\n".join(l for l in lines if ": FATAL" in l)
            raise BackendException(msg)
        elif is_failure:
            msg = "\n".join(l for l in lines if ": FAILURE" in l)
            raise BackendException(stderr)

    def ls(self, pattern="."):
        """
        List files on optical disc.
        """

        files = [f for f, _ in self.lsl(pattern)]

        return files

    def lsl(self, pattern="."):
        """
        List files on optical disc.
        """

        stdout, stderr = self.__send_cmd("-lsl", pattern)

        # Parse output of `xorriso -lsl`
        lines = stdout

        files = []

        for line in lines:
            line = line.strip()

            if not line:  # Skip empty lines
                continue

            parts = re.split("\\s+", line, maxsplit=8)

            if len(parts) != 9:
                continue

            mode, _, _uid, _gid, size, _month, _day, _time, filename = parts

            if not filename.startswith("'") or not filename.endswith("'"):
                raise BackendException("Got unexpected format from xorriso -lsl.")

            # Parse size into an integer.
            try:
                size = int(size)
            except Exception as e:
                raise BackendException("Could not parse file size.")

            filename = filename[1:-1]  # strip leading and trailing `'`s
            files.append((filename, {"size": size}))

        return files

    def commit(self):
        """
        Commit changes and write them to the image.
        """

        stdout, stderr = self.__send_cmd("-commit")

        if "exceeds free space on media" in "\n".join(stderr):
            raise BackendException("Not enough free space on media.")

    def end(self):
        """
        Terminate the xorriso subprocess
        """

        stdout, stderr = self.__send_cmd("-end")

    def cp(self, files, dest):
        """
        Copy file to the ISO image. Does not commit the changes yet.
        """
        assert isinstance(files, list), "should be a list of files"

        stdout, stderr = self.__send_cmd(
            "-cpr",
            *files,
            dest,
            "--",
        )

        if "exceeds free space on media" in "\n".join(stderr):
            raise BackendException("Not enough free space on media.")

    def rm(self, files):
        """
        Remove a list of files from the image. Does not commit the changes yet.
        """
        assert isinstance(files, list), "should be a list of files"

        if not files:
            return

        stdout, stderr = self.__send_cmd(
            "-rm",
            *files,
            # Don't commit yet.
        )

    def extract(self, files, dest):
        """
        Extract files from the ISO image.
        """
        assert isinstance(files, list), "should be a list of files"

        assert not os.path.exists(dest) or os.path.isfile(dest), "dest must be a directory"

        if len(files) == 0:
            return

        stdout, stderr = self.__send_cmd("-cpx", *files, dest)


class XorrisoBackend(duplicity.backend.Backend):
    """Backend for writing to optical discs or ISO images using xorriso.

    Simple URLs look like `xorriso:///dev/sr0` if the backup location is at the root of the filesystem.
    or if `xorriso://dev/sr0:/path/to/a/directory/on/iso` if the backup location is in a directory.

    Especially for testing also an ISO file can be used: xorriso://path/to/image.iso

    The path to the `xorriso` executable can be specified with teh `XORRISO_PATH` environment variable.
    Environment variables:

    * XORRISO_PATH: Alternative path to the `xorriso` executable
    * XORRISO_WRITE_SPEED: Specify the speed for writing to the optical disc. One of ["min", "max"].
    * XORRISO_ASSERT_VOLID: Abort when the volume ID of the ISO image does not match the given value.
    * XORRISO_ARGS: Arbitrary arguments to xorriso, inserted before the filesystem operations. For experts only.
    """

    def __init__(self, parsed_url):
        """
        Initialize the Xorriso backend.

        :param ParsedUrl parsed_url: The parsed URL.
        :raises InvalidBackendURL: If the URL is invalid.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        # Path to xorriso executable.
        xorriso_cmd = os.environ.get("XORRISO_PATH", default="xorriso")

        # Check if xorriso is installed.
        if xorriso_cmd == "xorriso":
            if not util.which("xorriso"):
                raise FatalBackendException("xorriso not installed")

        # Default arguments for xorriso.
        self.xorriso_args = []

        args_pre = os.environ.get("XORRISO_ARGS")
        if args_pre is not None:
            arg_list = shlex.split(args_pre)
            self.xorriso_args.extend(arg_list)

        assert_volid = os.environ.get("XORRISO_ASSERT_VOLID")
        if assert_volid is not None:
            self.xorriso_args += ["-assert_volid", assert_volid, "FAILURE"]

        speed = os.environ.get("XORRISO_WRITE_SPEED", default="min")
        if speed in ["min", "max"]:
            self.xorriso_args += ["-speed", speed]
        else:
            self.xorriso_args += ["-speed", "min"]

        # The URL form "file:MyFile" is not a valid duplicity target.
        if not parsed_url.path.startswith("//"):
            raise InvalidBackendURl("Bad xorriso:// path syntax.")

        path = parsed_url.path[2:]  # Strip '//'

        parts = path.split(":", maxsplit=1)
        if len(parts) == 2:
            self.device = parts[0]
            self.iso_path = parts[1]
        else:
            self.device = parts[0]
            self.iso_path = "/"

        if not self.iso_path.endswith("/"):
            self.iso_path += "/"

        if not os.path.exists(self.device):
            raise InvalidBackendURL(f"Optical disc device does not exist: {self.device}")

        # Start xorriso subprocess.
        self.xorriso = Xorriso(device=self.device, xorriso_path=xorriso_cmd, xorriso_args=self.xorriso_args)

    def _put(self, source_path, remote_filename):
        assert not os.path.isdir(source_path.name.decode("utf8")), "source_path must be a file"
        source_path.setdata()
        source_size = source_path.getsize()
        progress.report_transfer(0, source_size)

        self.xorriso.cp(
            [source_path.name.decode("utf8")],
            self.iso_path + remote_filename.decode("utf8"),
        )
        self.xorriso.commit()

        progress.report_transfer(source_size, source_size)

    def _get(self, filename, local_path):
        self.xorriso.extract([self.iso_path + filename.decode("utf8")], local_path.name.decode("utf8"))

    def _list(self):
        """
        List files in the ISO image.

        :returns: A list of filenames as byte strings.
        :rtype: list
        """
        files = self.xorriso.ls(pattern=self.iso_path)
        return [f.encode() for f in files]

    def _delete(self, filename):
        self.xorriso.rm([self.iso_path + filename.decode("utf8")])

    def _delete_list(self, filenames):
        filenames = [self.iso_path + f.decode("utf8") for f in filenames]
        self.xorriso.rm(filenames)

    def _query(self, filename):
        filename = self.iso_path + filename.decode("utf8")
        files = self.xorriso.lsl(filename)

        if len(files) == 0 or files[0][0] != filename:
            size = -1
        else:
            size = files[0][1]["size"]

        return {"size": size}

    def _close(self):
        self.xorriso.commit()
        self.xorriso.end()


duplicity.backend.register_backend("xorriso", XorrisoBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Generate and process backup statistics.
"""

import json
import os
import re
import time

from duplicity import config
from duplicity import dup_time
from duplicity import util
from duplicity import log


class StatsException(Exception):
    """
    Base exception for statistics-related errors.
    """

    pass


class StatsObj(object):
    """
    Contains various statistics, provide string conversion functions
    """

    # used when quoting files in get_stats_line
    space_regex = re.compile(" ")

    stat_file_attrs = (
        "SourceFiles",
        "SourceFileSize",
        "NewFiles",
        "NewFileSize",
        "DeletedFiles",
        "ChangedFiles",
        "ChangedFileSize",
        "ChangedDeltaSize",
        "DeltaEntries",
        "RawDeltaSize",
    )
    stat_misc_attrs = ("Errors", "TotalDestinationSizeChange")
    stat_time_attrs = ("StartTime", "EndTime", "ElapsedTime")
    stat_attrs = ("Filename",) + stat_time_attrs + stat_misc_attrs + stat_file_attrs

    # Below, the second value in each pair is true iff the value
    # indicates a number of bytes
    stat_file_pairs = (
        ("SourceFiles", False),
        ("SourceFileSize", True),
        ("NewFiles", False),
        ("NewFileSize", True),
        ("DeletedFiles", False),
        ("ChangedFiles", False),
        ("ChangedFileSize", True),
        ("ChangedDeltaSize", True),
        ("DeltaEntries", False),
        ("RawDeltaSize", True),
    )

    # This is used in get_byte_summary_string below
    byte_abbrev_list = (
        (1024 * 1024 * 1024 * 1024, "TB"),
        (1024 * 1024 * 1024, "GB"),
        (1024 * 1024, "MB"),
        (1024, "KB"),
    )

    def __init__(self):
        """
        Set attributes to None
        """
        for attr in self.stat_attrs:
            self.__dict__[attr] = None

    def get_stat(self, attribute):
        """
        Get a statistic
        """
        return self.__dict__[attribute]

    def set_stat(self, attr, value):
        """
        Set attribute to given value
        """
        self.__dict__[attr] = value

    def increment_stat(self, attr):
        """
        Add 1 to value of attribute
        """
        self.__dict__[attr] += 1

    def get_stats_line(self, index, use_repr=1):
        """
        Return one line abbreviated version of full stats string
        """
        file_attrs = [str(self.get_stat(a)) for a in self.stat_file_attrs]
        if not index:
            filename = "."
        else:
            filename = os.path.join(*index)
            if use_repr:
                # use repr to quote newlines in relative filename, then
                # take of leading and trailing quote and quote spaces.
                filename = self.space_regex.sub("\\\\x20", repr(filename))
                n = 1
                if filename[0] == "u":
                    n = 2
                filename = filename[n:-1]
        return " ".join(
            [
                filename,
            ]
            + file_attrs
        )

    def set_stats_from_line(self, line):
        """
        Set statistics from given line
        """

        def error():
            raise StatsException(f"Bad line '{line}'")

        if line[-1] == "\n":
            line = line[:-1]
        lineparts = line.split(" ")
        if len(lineparts) < len(self.stat_file_attrs):
            error()
        for attr, val_string in zip(self.stat_file_attrs, lineparts[-len(self.stat_file_attrs) :]):
            try:
                val = int(val_string)
            except ValueError:
                try:
                    val = float(val_string)
                except ValueError:
                    error()
            self.set_stat(attr, val)
        return self

    def get_stats_string(self):
        """
        Return extended string printing out statistics
        """
        return f"{self.get_timestats_string()}{self.get_filestats_string()}{self.get_miscstats_string()}"

    def get_timestats_string(self):
        """
        Return portion of statistics string dealing with time
        """
        timelist = []
        if self.StartTime is not None:
            timelist.append(
                "StartTime %.2f (%s)\n"
                % (  # pylint: disable=bad-string-format-type
                    self.StartTime,
                    dup_time.timetopretty(self.StartTime),
                )
            )
        if self.EndTime is not None:
            timelist.append(
                "EndTime %.2f (%s)\n"
                % (  # pylint: disable=bad-string-format-type
                    self.EndTime,
                    dup_time.timetopretty(self.EndTime),
                )
            )
        if self.ElapsedTime or (  # pylint:disable=access-member-before-definition
            self.StartTime is not None and self.EndTime is not None
        ):
            if self.ElapsedTime is None:  # pylint:disable=access-member-before-definition
                self.ElapsedTime = self.EndTime - self.StartTime
            timelist.append(f"ElapsedTime {self.ElapsedTime:.2f} ({dup_time.inttopretty(self.ElapsedTime)})\n")
        return "".join(timelist)

    def get_filestats_string(self):
        """
        Return portion of statistics string about files and bytes
        """

        def fileline(stat_file_pair):
            """
            Return zero or one line of the string
            """
            attr, in_bytes = stat_file_pair
            val = self.get_stat(attr)
            if val is None:
                return ""
            if in_bytes:
                return f"{attr} {val} ({self.get_byte_summary_string(val)})\n"
            else:
                return f"{attr} {val}\n"

        return "".join(map(fileline, self.stat_file_pairs))

    def get_miscstats_string(self):
        """
        Return portion of extended stat string about misc attributes
        """
        misc_string = ""
        tdsc = self.TotalDestinationSizeChange
        if tdsc is not None:
            misc_string += f"TotalDestinationSizeChange {tdsc} ({self.get_byte_summary_string(tdsc)})\n"
        if self.Errors is not None:
            misc_string += f"Errors {int(self.Errors)}\n"
        return misc_string

    def get_byte_summary_string(self, byte_count):
        """
        Turn byte count into human readable string like "7.23GB"
        """
        if byte_count < 0:
            sign = "-"
            byte_count = -byte_count
        else:
            sign = ""

        for abbrev_bytes, abbrev_string in self.byte_abbrev_list:
            if byte_count >= abbrev_bytes:
                # Now get 3 significant figures
                abbrev_count = float(byte_count) / abbrev_bytes
                if abbrev_count >= 100:
                    precision = 0
                elif abbrev_count >= 10:
                    precision = 1
                else:
                    precision = 2
                return f"{sign}{abbrev_count:.{precision}f} {abbrev_string}"
        byte_count = round(byte_count)
        if byte_count == 1:
            return f"{sign}1 byte"
        else:
            return f"{sign}{int(byte_count)} bytes"

    def get_stats_json(self, col_stat):
        """
        Return enriched statistics in JSON format.

        :param col_stat: The collection status, used to gather information
            about the whole backup chain.
        :returns: JSON formatted string.
        :rtype: str
        """

        def fail_save_read(parent, *attributes, is_function=False, default="N/A"):
            """
            Return the value of an attribute, or a default if it cannot be determined.

            :param object parent: The object to read from.
            :param str attributes: Path down to the attribute that should be read.
            :param bool is_function: Whether the target attribute is a function to be called.
            :param default: The default value to return on failure.
            :returns: The attribute value or the default.
            """
            try:
                attr_path = parent
                for attribute in attributes:
                    try:
                        attr_path = getattr(attr_path, attribute)
                    except AttributeError:
                        return default
                if is_function:
                    return attr_path()
                else:
                    return attr_path
            except Exception as e:
                log.Error(f"Can't read expected attribute: {e}")
                return default

        py_obj = {key: self.__dict__[key] for key in self.stat_attrs}
        for t in ("StartTime", "EndTime"):
            t_str = f"{t}_str"
            py_obj[t_str] = dup_time.timetostring(py_obj[t])
        if not py_obj.get("ElapsedTime"):
            py_obj["ElapsedTime"] = py_obj["EndTime"] - py_obj["StartTime"]
        if col_stat:
            backup_meta = {}
            backup_chain = col_stat.matched_chain_pair[1]
            backup_meta["action"] = config.action
            backup_meta["skipped_inc"] = config.skipped_inc
            backup_meta["time_full_bkp"] = backup_chain.fullset.time
            backup_meta["time_full_bkp_str"] = dup_time.timetostring(backup_meta["time_full_bkp"])
            backup_meta["no_of_inc"] = len(backup_chain.incset_list)
            backup_meta["concurrency"] = fail_save_read(config, "concurrency")
            backup_meta["target"] = fail_save_read(config, "target_url")
            backup_meta["source"] = fail_save_read(config, "source_path")
            backup_meta["local_json_stat"] = [
                fail_save_read(
                    backup_chain,
                    "fullset",
                    "local_jsonstat_path",
                    "get_filename",
                    is_function=True,
                )
            ]
            for inc in backup_chain.incset_list:
                backup_meta["local_json_stat"].append(
                    fail_save_read(inc, "local_jsonstat_path", "get_filename", is_function=True)
                )
            py_obj["backup_meta"] = backup_meta

        return json.dumps(py_obj, cls=util.BytesEncoder, indent=4)

    def get_stats_logstring(self, title):
        """
        Like get_stats_string, but add header and footer
        """
        header = f"--------------[ {title} ]--------------"
        footer = "-" * len(header)
        return f"{header}\n{self.get_stats_string()}{footer}\n"

    def set_stats_from_string(self, s):
        """
        Initialize attributes from string, return self for convenience
        """

        def error(line):
            raise StatsException(f"Bad line '{line}'")

        for line in s.split("\n"):
            if not line:
                continue
            line_parts = line.split()
            if len(line_parts) < 2:
                error(line)
            attr, value_string = line_parts[:2]
            if attr not in self.stat_attrs:
                error(line)
            try:
                try:
                    val1 = int(value_string)
                except ValueError:
                    val1 = None
                val2 = float(value_string)
                if val1 == val2:
                    self.set_stat(attr, val1)  # use integer val
                else:
                    self.set_stat(attr, val2)  # use float
            except ValueError:
                error(line)
        return self

    def write_stats_to_path(self, path):
        """
        Write statistics string to given path
        """
        fin = path.open("w")
        fin.write(self.get_stats_string())
        assert not fin.close(), "fin failed to close"

    def read_stats_from_path(self, path):
        """
        Set statistics from path, return self for convenience
        """
        fp = path.open("r")
        self.set_stats_from_string(fp.read())
        assert not fp.close(), "fp failed to close"
        return self

    def stats_equal(self, s):
        """
        Return true if s has same statistics as self
        """
        assert isinstance(s, StatsObj), "Comparison target must be a StatsObj instance"
        for attr in self.stat_file_attrs:
            if self.get_stat(attr) != s.get_stat(attr):
                return None
        return 1

    def set_to_average(self, statobj_list):
        """
        Set self's attributes to average of those in statobj_list
        """
        for attr in self.stat_attrs:
            self.set_stat(attr, 0)
        for statobj in statobj_list:
            for attr in self.stat_attrs:
                if statobj.get_stat(attr) is None:
                    self.set_stat(attr, None)
                elif self.get_stat(attr) is not None:
                    self.set_stat(attr, statobj.get_stat(attr) + self.get_stat(attr))

        # Don't compute average starting/stopping time
        self.StartTime = None
        self.EndTime = None

        for attr in self.stat_attrs:
            if self.get_stat(attr) is not None:
                self.set_stat(attr, self.get_stat(attr) / float(len(statobj_list)))
        return self

    def get_statsobj_copy(self):
        """
        Return new StatsObj object with same stats as self
        """
        s = StatsObj()
        for attr in self.stat_attrs:
            s.set_stat(attr, self.get_stat(attr))
        return s


class StatsDeltaProcess(StatsObj):
    """
    Keep track of statistics during DirDelta process
    """

    def __init__(self):
        """
        StatsDeltaProcess initializer - zero file attributes
        """
        StatsObj.__init__(self)
        for attr in StatsObj.stat_file_attrs:
            self.__dict__[attr] = 0
        self.Errors = 0
        self.StartTime = time.time()
        self.files_changed = []

    def add_new_file(self, path):
        """
        Add stats of new file path to statistics
        """
        filesize = path.getsize()
        self.SourceFiles += 1
        # SourceFileSize is added-to incrementally as read
        self.NewFiles += 1
        self.NewFileSize += filesize
        self.DeltaEntries += 1
        self.add_delta_entries_file(path, b"new")

    def add_changed_file(self, path):
        """
        Add stats of file that has changed since last backup
        """
        filesize = path.getsize()
        self.SourceFiles += 1
        # SourceFileSize is added-to incrementally as read
        self.ChangedFiles += 1
        self.ChangedFileSize += filesize
        self.DeltaEntries += 1
        self.add_delta_entries_file(path, b"changed")

    def add_deleted_file(self, path):
        """
        Add stats of file no longer in source directory
        """
        self.DeletedFiles += 1  # can't add size since not available
        self.DeltaEntries += 1
        self.add_delta_entries_file(path, b"deleted")

    def add_unchanged_file(self, path):
        """
        Add stats of file that hasn't changed since last backup
        """
        filesize = path.getsize()
        self.SourceFiles += 1
        self.SourceFileSize += filesize

    def close(self):
        """
        End collection of data, set EndTime
        """
        self.EndTime = time.time()

    def add_delta_entries_file(self, path, action_type):
        if config.files_changed and path.isreg():
            self.files_changed.append((path.get_relative_path(), action_type))

    def get_delta_entries_file(self):
        return self.files_changed



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2017 Tomas Vondra
# Copyright 2017 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Mega.co.nz backend using megatools.
"""

import os
import duplicity.backend
from duplicity.errors import BackendException


class MegaBackend(duplicity.backend.Backend):
    """
    Backend for Mega.co.nz using megatools.
    """

    def __init__(self, parsed_url):
        """
        Initialize the Mega backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        # ensure all the necessary megatools binaries exist
        self._check_binary_exists("megals")
        self._check_binary_exists("megamkdir")
        self._check_binary_exists("megaget")
        self._check_binary_exists("megaput")
        self._check_binary_exists("megarm")

        # store some basic info
        self._hostname = parsed_url.hostname

        if parsed_url.password is None:
            self._megarc = f"{os.getenv('HOME')}/.megarc"
        else:
            self._megarc = False
            self._username = parsed_url.username
            self._password = self.get_password()

        # remote folder (Can we assume /Root prefix?)
        self._root = "/Root"
        self._folder = f"{self._root}/{parsed_url.path[1:]}"

        # make sure the remote folder exists (the whole path)
        self._makedir_recursive(parsed_url.path[1:].split("/"))

    def _check_binary_exists(self, cmd):
        """
        checks that a specified command exists in the current path
        """

        try:
            # ignore the output, we only need the return code
            self.subprocess_popen(["which", cmd])
        except Exception as e:
            raise BackendException(f"command '{cmd}' not found, make sure megatools are installed")

    def _makedir(self, path):
        """
        creates a remote directory
        """

        if self._megarc:
            cmd = ["megamkdir", "--config", self._megarc, path]
        else:
            cmd = ["megamkdir", "-u", self._username, "-p", self._password, path]

        self.subprocess_popen(cmd)

    def _makedir_recursive(self, path):
        """
        creates a remote directory (recursively the whole path), ingores errors
        """

        print(f"mkdir: {'/'.join(path)}")

        p = self._root

        for folder in path:
            p = f"{p}/{folder}"
            try:
                self._makedir(p)
            except Exception as e:
                pass

    def _put(self, source_path, remote_filename):
        """
        uploads file to Mega (deletes it first, to ensure it does not exist)
        """

        try:
            self.delete(os.fsdecode(remote_filename))
        except Exception:
            pass

        self.upload(local_file=os.fsdecode(source_path.get_canonical()), remote_file=os.fsdecode(remote_filename))

    def _get(self, remote_filename, local_path):
        """
        downloads file from Mega
        """

        self.download(remote_file=os.fsdecode(remote_filename), local_file=os.fsdecode(local_path.name))

    def _list(self):
        """
        list files in the backup folder
        """

        return self.folder_contents(files_only=True)

    def _delete(self, filename):
        """
        deletes remote
        """

        self.delete(remote_file=os.fsdecode(filename))

    def folder_contents(self, files_only=False):
        """
        lists contents of a folder, optionally ignoring subdirectories
        """

        print(f"megals: {self._folder}")

        if self._megarc:
            cmd = ["megals", "--config", self._megarc, self._folder]
        else:
            cmd = ["megals", "-u", self._username, "-p", self._password, self._folder]

        result, stdout, stderr = self.subprocess_popen(cmd)
        files = os.fsdecode(stdout.strip()).split("\n")

        # remove the folder name, including the path separator
        files = [f[len(self._folder) + 1 :] for f in files]

        # optionally ignore entries containing path separator (i.e. not files)
        if files_only:
            files = [f for f in files if "/" not in f]

        return [os.fsencode(f) for f in files]

    def download(self, remote_file, local_file):
        print(f"megaget: {remote_file}")

        if self._megarc:
            cmd = [
                "megaget",
                "--config",
                self._megarc,
                "--no-progress",
                "--path",
                local_file,
                f"{self._folder}/{remote_file}",
            ]
        else:
            cmd = [
                "megaget",
                "-u",
                self._username,
                "-p",
                self._password,
                "--no-progress",
                "--path",
                local_file,
                f"{self._folder}/{remote_file}",
            ]

        self.subprocess_popen(cmd)

    def upload(self, local_file, remote_file):
        print(f"megaput: {remote_file}")

        if self._megarc:
            cmd = [
                "megaput",
                "--config",
                self._megarc,
                "--no-progress",
                "--path",
                f"{self._folder}/{remote_file}",
                local_file,
            ]
        else:
            cmd = [
                "megaput",
                "-u",
                self._username,
                "-p",
                self._password,
                "--no-progress",
                "--path",
                f"{self._folder}/{remote_file}",
                local_file,
            ]

        try:
            self.subprocess_popen(cmd)
        except Exception as e:
            error_str = str(e)
            if "EOVERQUOTA" in error_str:
                raise BackendException(
                    f"MEGA account over quota, could not write file : '{remote_file}'. "
                    f"Upgrade your storage at https://mega.nz/pro or remove some data."
                )
            else:
                raise BackendException(f"Failed writing file '{remote_file}' to MEGA , reason : '{e}'")

    def delete(self, remote_file):
        print(f"megarm: {remote_file}")

        if self._megarc:
            cmd = ["megarm", "--config", self._megarc, f"{self._folder}/{remote_file}"]
        else:
            cmd = ["megarm", "-u", self._username, "-p", self._password, f"{self._folder}/{remote_file}"]

        self.subprocess_popen(cmd)


duplicity.backend.register_backend("mega", MegaBackend)
duplicity.backend.uses_netloc.extend(["mega"])



#!/usr/bin/env python3
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# duplicity -- Encrypted bandwidth efficient backup
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# See http://www.nongnu.org/duplicity for more information.
# Please send mail to me or the mailing list if you find bugs or have
# any suggestions.

"""
Main entry point for Duplicity.
"""

import errno
import os
import sys

from duplicity import errors
from duplicity import log
from duplicity import tempdir
from duplicity import util
from duplicity.dup_main import main
from duplicity.gpg import GPGError

sys.stdout.reconfigure(errors="surrogateescape")
sys.stderr.reconfigure(errors="surrogateescape")


def with_tempdir(fn):
    """
    Execute function and guarantee cleanup of tempdir is called.

    :param callable fn: The function to execute.
    """
    try:
        fn()
    finally:
        tempdir.default().cleanup()


def dup_run():
    """
    Main function for running Duplicity.

    Handles initialization, command line processing, and exception handling.
    """
    # check that we can function here
    if os.environ.get("PYTEST_VERSION") is not None:
        pass
    elif os.environ.get("CIBUILDWHEEL") is not None:
        pass
    elif not ((3, 10) <= sys.version_info[:2] <= (3, 14)):
        print("Sorry, duplicity requires Python version 3.10 thru 3.14.", file=sys.stderr)
        sys.exit(1)

    try:
        try:
            log.setup()
            util.start_debugger()
            with_tempdir(main)

        # Don't move this lower.  To get an exit status out of the system, you have to call
        # the sys.exit() function.  Python handles this by raising the SystemExit exception.
        # Cleanup code goes here, if needed.
        except SystemExit as e:
            # No traceback, just get out
            util.release_lockfile()
            sys.exit(e.code)

        except KeyboardInterrupt as e:
            util.release_lockfile()
            log.Info(_("INT intercepted...exiting."))
            sys.exit(errno.EINTR)

        except GPGError as e:
            log.Info(_("GPG error detail: %s") % util.exception_traceback())
            raise errors.FatalError(
                f"{e.__class__.__name__}: {e.args[0]}", log.ErrorCode.gpg_failed, e.__class__.__name__
            )

        except errors.UserError as e:
            log.Info(_("User error detail: %s") % util.exception_traceback())
            raise errors.FatalError(
                f"{e.__class__.__name__}: {util.uexc(e)}", log.ErrorCode.user_error, e.__class__.__name__
            )

        except errors.BackendException as e:
            log.Info(_("Backend error detail: %s") % util.exception_traceback())
            raise errors.FatalError(
                f"{e.__class__.__name__}: {util.uexc(e)}", log.ErrorCode.backend_error, e.__class__.__name__
            )

        except errors.FatalError:
            raise

        except Exception as e:
            if "Forced assertion for testing" in util.uexc(e):
                raise errors.FatalError(
                    f"{e.__class__.__name__}: {util.uexc(e)}", log.ErrorCode.exception, e.__class__.__name__
                )
            else:
                raise errors.FatalError(util.exception_traceback(), log.ErrorCode.exception, e.__class__.__name__)

    except errors.FatalError as e:
        util.release_lockfile()
        log.shutdown()
        sys.exit(e.code)


if __name__ == "__main__":
    dup_run()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2020 Jose L. Domingo Lopez
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
MEGA.nz backend using MEGAcmd (V3).
"""

import os
import re
import duplicity.backend
from duplicity.errors import BackendException


class Megav3Backend(duplicity.backend.Backend):
    """Backend for MEGA.nz cloud storage, only one that works for accounts created since Nov. 2018
    See https://github.com/megous/megatools/issues/411 for more details

    This MEGA backend resorts to official tools (MEGAcmd) as available at https://mega.nz/cmd
    MEGAcmd works through a single binary called "mega-cmd", which keeps state (for example,
    persisting a session). Multiple ``mega-*`` shell wrappers (ie. "mega-ls") exist as the user
    interface to "mega-cmd" and MEGA API
    The full MEGAcmd User Guide can be found in the software's GitHub page below :
    https://github.com/meganz/MEGAcmd/blob/master/UserGuide.md"""

    def __init__(self, parsed_url):
        """
        Initialize the MEGA backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        # Sanity check : ensure all the necessary "MEGAcmd" binaries exist
        self._check_binary_exists("mega-cmd")
        self._check_binary_exists("mega-exec")
        self._check_binary_exists("mega-help")
        self._check_binary_exists("mega-get")
        self._check_binary_exists("mega-login")
        self._check_binary_exists("mega-logout")
        self._check_binary_exists("mega-ls")
        self._check_binary_exists("mega-mkdir")
        self._check_binary_exists("mega-put")
        self._check_binary_exists("mega-rm")
        self._check_binary_exists("mega-whoami")

        # "MEGAcmd" does not use a config file, however it is handy to keep one (with the old ".megarc" format) to
        # securely store the username and password
        self._hostname = parsed_url.hostname
        if parsed_url.username is None:
            self._megarc = f"{os.getenv('HOME')}/.megav3rc"
            try:
                conf_file = open(self._megarc, "r")
            except Exception as e:
                raise BackendException(
                    f"No password provided in URL and MEGA configuration file for "
                    f"duplicity does not exist as '{self._megarc}'"
                )

            myvars = {}
            for line in conf_file:
                name, var = line.partition("=")[::2]
                myvars[name.strip()] = str(var.strip())
            conf_file.close()
            self._username = myvars["Username"]
            self._password = myvars["Password"]

        else:
            self._username = parsed_url.username
            self._password = parsed_url.password

        no_logout_option = parsed_url.query_args.get("no_logout", [])
        self._no_logout = (len(no_logout_option) > 0) and (no_logout_option[0].lower() in ["1", "yes", "true"])

        self.ensure_mega_cmd_running()

        # Remote folder ("MEGAcmd" no longer shows "Root/" at the top of the hierarchy)
        self._folder = f"/{parsed_url.path[1:]}"

        # Only create the remote folder if it doesn't exist yet
        self.mega_login()
        cmd = ["mega-ls", self._folder]
        try:
            self.subprocess_popen(cmd)
        except Exception as e:
            self._makedir(self._folder)

    def _check_binary_exists(self, cmd):
        """
        Checks that a specified command exists in the running user command path
        """

        try:
            # Ignore the output, as we only need the return code
            self.subprocess_popen(["which", cmd])
        except Exception as e:
            raise BackendException(
                f"Command '{cmd}' not found, make sure 'MEGAcmd' tools (https://mega.nz/cmd) is properly installed "
                f"and in the running user command path"
            )

    def ensure_mega_cmd_running(self):
        """
        Trigger any mega command to ensure mega-cmd server is running
        """
        try:
            self.subprocess_popen("mega-help")
        except Exception:
            raise BackendException("Cannot execute mega command")

    def _makedir(self, path):
        """
        Creates a remote directory (recursively if necessary)
        """

        self.mega_login()
        cmd = ["mega-mkdir", "-p", path]
        try:
            self.subprocess_popen(cmd)
        except Exception as e:
            error_str = str(e)
            if "Folder already exists" in error_str:
                raise BackendException(
                    f"Folder '{path}' could not be created on MEGA because it already exists. "
                    f"Use another path or remove the folder in MEGA manually"
                )
            else:
                raise BackendException(f"Folder '{path}' could not be created, reason : '{e}'")

    def _put(self, source_path, remote_filename):
        """Uploads file to the specified remote folder (tries to delete it first to make
        sure the new one can be uploaded)"""

        try:
            self.delete(remote_filename.decode())
        except Exception:
            pass
        self.upload(
            local_file=source_path.get_canonical().decode(),
            remote_file=remote_filename.decode(),
        )

    def _get(self, remote_filename, local_path):
        """
        Downloads file from the specified remote path
        """

        self.download(
            remote_file=remote_filename.decode(),
            local_file=local_path.name.decode(),
        )

    def _list(self):
        """
        Lists files in the specified remote path
        """

        return self.folder_contents(files_only=True)

    def _delete(self, filename):
        """
        Deletes file from the specified remote path
        """

        self.delete(remote_file=filename.decode())

    def _close(self):
        """
        Function called when backend is done being used
        """

        if not self._no_logout:
            cmd = ["mega-logout"]
            self.subprocess_popen(cmd)

        cmd = ["mega-exec", "exit"]
        self.subprocess_popen(cmd)

    def mega_login(self):
        """
        Helper function to check existing session exists
        """

        # Abort if command doesn't return in a reasonable time (somehow "mega-session" sometimes
        # doesn't return), and create session if one doesn't exist yet
        try:
            result, stdout, stderr = self.subprocess_popen("mega-whoami", timeout=30)
            current_username = stdout.split(":")[-1].strip()
            if current_username != self._username:
                raise Exception("Username is not match")
        except Exception as e:
            if "Timed out" in str(e):
                self._close()
                raise BackendException("Timed out while trying to determine if a MEGA session exists")

            if self._password is None:
                self._password = self.get_password()

            cmd = ["mega-login", self._username, self._password]
            try:
                self.subprocess_popen(cmd)
            except Exception as e:
                self._close()
                raise BackendException(f"Could not log in to MEGA, error : '{e}'")

    def folder_contents(self, files_only=False):
        """
        Lists contents of a remote MEGA path, optionally ignoring subdirectories
        """

        cmd = ["mega-ls", "-l", self._folder]

        self.mega_login()
        result, stdout, stderr = self.subprocess_popen(cmd)
        files = stdout.split("\n")

        # Optionally ignore directories
        if files_only:
            files = [f.split()[5] for f in files if re.search("^-", f)]

        return files

    def download(self, remote_file, local_file):
        """
        Downloads a file from a remote MEGA path
        """

        cmd = ["mega-get", f"{self._folder}/{remote_file}", local_file]
        self.mega_login()
        self.subprocess_popen(cmd)

    def upload(self, local_file, remote_file):
        """
        Uploads a file to a remote MEGA path
        """

        cmd = ["mega-put", local_file, f"{self._folder}/{remote_file}"]
        self.mega_login()
        try:
            self.subprocess_popen(cmd)
        except Exception as e:
            error_str = str(e)
            if "Reached storage quota" in error_str:
                raise BackendException(
                    f"MEGA account over quota, could not write file : '{remote_file}' . "
                    f"Upgrade your storage at https://mega.nz/pro or remove some data."
                )
            else:
                raise BackendException(f"Failed writing file '{remote_file}' to MEGA, reason : '{e}'")

    def delete(self, remote_file):
        """
        Deletes a file from a remote MEGA path
        """

        cmd = ["mega-rm", "-f", f"{self._folder}/{remote_file}"]
        self.mega_login()
        self.subprocess_popen(cmd)


duplicity.backend.register_backend("megav3", Megav3Backend)
duplicity.backend.uses_netloc.extend(["megav3"])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""Wrapper class around a file like "/usr/bin/env"

This class makes certain file operations more convenient and
associates stat information with filenames

"""

import errno
import gzip
import os
import re
import shutil
import socket
import stat
import subprocess
import time

from duplicity import cached_ops
from duplicity import config
from duplicity import dup_tarfile
from duplicity import dup_time
from duplicity import file_naming
from duplicity import gpg
from duplicity import log
from duplicity import util
from duplicity.lazy import Iter

_tmp_path_counter = 1


class StatResult(object):
    """
    Used to emulate the output of os.stat() and related
    """

    # st_mode is required by the TarInfo class, but it's unclear how
    # to generate it from file permissions.
    st_mode = 0


class PathException(Exception):
    """
    Base exception for path errors.
    """

    pass


class ROPath(object):
    """
    Read-only Path.

    Objects of this class don't represent real files on disk, so they don't
    have a name. They are required to be indexed, however.
    """

    def __init__(self, index, stat=None):  # pylint: disable=unused-argument
        """
        Initialize an ROPath.

        :param tuple index: The index of the path.
        :param stat: The stat information for the path.
        """
        self.opened, self.fileobj = None, None
        self.index = index
        self.stat, self.type = None, None
        self.mode, self.devnums = None, None

    def set_from_stat(self):
        """
        Set the values of ``self.type`` and ``self.mode`` from ``self.stat``.

        :raises PathException: If the file type is unknown or unsupported (e.g. sockets).
        """
        if not self.stat:
            self.type = None

        st_mode = self.stat.st_mode
        if stat.S_ISREG(st_mode):
            self.type = "reg"
        elif stat.S_ISDIR(st_mode):
            self.type = "dir"
        elif stat.S_ISLNK(st_mode):
            self.type = "sym"
        elif stat.S_ISFIFO(st_mode):
            self.type = "fifo"
        elif stat.S_ISSOCK(st_mode):
            raise PathException(f"{os.fsdecode(self.get_relative_path())}is a socket, unsupported by tar")
            self.type = "sock"  # pylint: disable=unreachable
        elif stat.S_ISCHR(st_mode):
            self.type = "chr"
        elif stat.S_ISBLK(st_mode):
            self.type = "blk"
        else:
            raise PathException("Unknown type")

        self.mode = stat.S_IMODE(st_mode)
        if self.type in ("chr", "blk"):
            try:
                self.devnums = (
                    os.major(self.stat.st_rdev),
                    os.minor(self.stat.st_rdev),
                )
            except Exception as e:
                log.Warn(
                    _("WARNING: %s invalid devnums (0x%X), treating as (0, 0).")
                    % (os.fsdecode(self.get_relative_path()), self.stat.st_rdev)
                )
                self.devnums = (0, 0)

    def blank(self):
        """
        Black out self - set type and stat to None
        """
        self.type, self.stat = None, None

    def exists(self):
        """
        True if corresponding file exists
        """
        return self.type

    def isreg(self):
        """
        True if self corresponds to regular file
        """
        return self.type == "reg"

    def isdir(self):
        """
        True if self is dir
        """
        return self.type == "dir"

    def issym(self):
        """
        True if self is sym
        """
        return self.type == "sym"

    def isfifo(self):
        """
        True if self is fifo
        """
        return self.type == "fifo"

    def issock(self):
        """
        True is self is socket
        """
        return self.type == "sock"

    def isdev(self):
        """
        True is self is a device file
        """
        return self.type == "chr" or self.type == "blk"

    def getdevloc(self):
        """
        Return device number path resides on
        """
        return self.stat.st_dev

    def getsize(self):
        """
        Return length in bytes from stat object
        """
        return self.stat.st_size

    def getmtime(self):
        """
        Return mod time of path in seconds
        """
        return int(self.stat.st_mtime)

    def get_relative_path(self):
        """
        Return relative path, created from index
        """
        if self.index:
            return b"/".join(self.index)
        else:
            return b"."

    def getperms(self):
        """
        Return permissions mode, owner and group
        """
        s1 = self.stat
        return f"{s1.st_uid}:{s1.st_gid} {self.mode:o}"

    def open(self, mode):
        """
        Return fileobj associated with self
        """
        assert mode == "rb" and self.fileobj and not self.opened, f"{mode} {self.fileobj} {self.opened}"
        self.opened = 1
        return self.fileobj

    def get_data(self):
        """
        Return contents of associated fileobj in string
        """
        fin = self.open("rb")
        buf = fin.read()
        assert not fin.close(), "fin failed to close"
        return buf

    def setfileobj(self, fileobj):
        """
        Set file object returned by open()
        """
        assert not self.fileobj, "fileobj is already set"
        self.fileobj = fileobj
        self.opened = None

    def init_from_tarinfo(self, tarinfo):
        """
        Set data from tarinfo object (part of dup_tarfile module)
        """
        # Set the typepp
        type = tarinfo.type  # pylint: disable=redefined-builtin
        if type == dup_tarfile.REGTYPE or type == dup_tarfile.AREGTYPE:
            self.type = "reg"
        elif type == dup_tarfile.LNKTYPE:
            raise PathException("Hard links not supported yet")
        elif type == dup_tarfile.SYMTYPE:
            self.type = "sym"
            self.symtext = tarinfo.linkname
            if isinstance(self.symtext, str):
                self.symtext = os.fsencode(self.symtext)
        elif type == dup_tarfile.CHRTYPE:
            self.type = "chr"
            self.devnums = (tarinfo.devmajor, tarinfo.devminor)
        elif type == dup_tarfile.BLKTYPE:
            self.type = "blk"
            self.devnums = (tarinfo.devmajor, tarinfo.devminor)
        elif type == dup_tarfile.DIRTYPE:
            self.type = "dir"
        elif type == dup_tarfile.FIFOTYPE:
            self.type = "fifo"
        else:
            raise PathException(f"Unknown tarinfo type {type}")

        self.mode = tarinfo.mode
        self.stat = StatResult()

        """ If do_not_restore_owner is False,
        set user and group id
        use numeric id if name lookup fails
        OR
        --numeric-owner is set
        """
        if config.numeric_owner or (passwd := cached_ops.getpwnam(tarinfo.uname)) is None:
            self.stat.st_uid = tarinfo.uid
        else:
            self.stat.st_uid = passwd[2]
        if config.numeric_owner or (group := cached_ops.getgrnam(tarinfo.gname)) is None:
            self.stat.st_gid = tarinfo.gid
        else:
            self.stat.st_gid = group[2]

        self.stat.st_mtime = int(tarinfo.mtime)
        if self.stat.st_mtime < 0:
            log.Warn(_("WARNING: %s has negative mtime, treating as 0.") % tarinfo.uc_name)
            self.stat.st_mtime = 0
        self.stat.st_size = tarinfo.size

    def get_ropath(self):
        """
        Return ropath copy of self
        """
        new_ropath = ROPath(self.index, self.stat)
        new_ropath.type, new_ropath.mode = self.type, self.mode
        if self.issym():
            new_ropath.symtext = self.symtext
        elif self.isdev():
            new_ropath.devnums = self.devnums
        if self.exists():
            new_ropath.stat = self.stat
        return new_ropath

    def get_tarinfo(self):
        """Generate a dup_tarfile.TarInfo object based on self

        Doesn't set size based on stat, because we may want to replace
        data wiht other stream.  Size should be set separately by
        calling function.

        """
        ti = dup_tarfile.TarInfo()
        if self.index:
            ti.name = os.fsdecode(b"/".join(self.index))
        else:
            ti.name = "."
        if self.isdir():
            ti.name += "/"  # tar dir naming convention

        ti.size = 0
        if self.type:
            # Lots of this is specific to dup_tarfile.py, hope it doesn't
            # change much...
            if self.isreg():
                ti.type = dup_tarfile.REGTYPE
                ti.size = self.stat.st_size
            elif self.isdir():
                ti.type = dup_tarfile.DIRTYPE
            elif self.isfifo():
                ti.type = dup_tarfile.FIFOTYPE
            elif self.issym():
                ti.type = dup_tarfile.SYMTYPE
                ti.linkname = self.symtext
                if isinstance(ti.linkname, bytes):
                    ti.linkname = os.fsdecode(ti.linkname)
            elif self.isdev():
                if self.type == "chr":
                    ti.type = dup_tarfile.CHRTYPE
                else:
                    ti.type = dup_tarfile.BLKTYPE
                ti.devmajor, ti.devminor = self.devnums
            else:
                raise PathException(f"Unrecognized type {str(self.type)}")

            ti.mode = self.mode
            ti.uid, ti.gid = self.stat.st_uid, self.stat.st_gid
            if self.stat.st_mtime < 0:
                log.Warn(_("WARNING: %s has negative mtime, treating as 0.") % (os.fsdecode(self.get_relative_path())))
                ti.mtime = 0
            else:
                ti.mtime = int(self.stat.st_mtime)

            try:
                ti.uname = cached_ops.getpwuid(ti.uid)[0]
            except KeyError:
                ti.uname = ""
            try:
                ti.gname = cached_ops.getgrgid(ti.gid)[0]
            except KeyError:
                ti.gname = ""

            if ti.type in (dup_tarfile.CHRTYPE, dup_tarfile.BLKTYPE):
                if hasattr(os, "major") and hasattr(os, "minor"):
                    ti.devmajor, ti.devminor = self.devnums
        else:
            # Currently we depend on an uninitiliazed tarinfo file to
            # already have appropriate headers.  Still, might as well
            # make sure mode and size set.
            ti.mode, ti.size = 0, 0
        return ti

    def __eq__(self, other):
        """
        Used to compare two ROPaths.  Doesn't look at fileobjs
        """
        if not self.type and not other.type:
            return 1  # neither exists
        if not self.stat and other.stat or not other.stat and self.stat:
            return 0
        if self.type != other.type:
            return 0

        if self.isreg() or self.isdir() or self.isfifo():
            # Don't compare sizes, because we might be comparing
            # signature size to size of file.
            if not self.perms_equal(other):
                return 0
            if int(self.stat.st_mtime) == int(other.stat.st_mtime):
                return 1
            # Below, treat negative mtimes as equal to 0
            return self.stat.st_mtime <= 0 and other.stat.st_mtime <= 0
        elif self.issym():
            # here only symtext matters
            return self.symtext == other.symtext
        elif self.isdev():
            return self.perms_equal(other) and self.devnums == other.devnums
        assert 0, "Unhandled path type in ROPath.__eq__"

    def __ne__(self, other):
        return not self.__eq__(other)

    def compare_verbose(self, other, include_data=0):
        """Compare ROPaths like __eq__, but log reason if different

        This is placed in a separate function from __eq__ because
        __eq__ should be very time sensitive, and logging statements
        would slow it down.  Used when verifying.

        Only run if include_data is true.

        """

        def log_diff(log_string):
            log_str = f"{_('Difference found:')} {log_string}"
            log.Notice(log_str % (os.fsdecode(self.get_relative_path())))

        if include_data is False:
            return True

        if not self.type and not other.type:
            return 1
        if not self.stat and other.stat:
            log_diff(_("New file %s"))
            return 0
        if not other.stat and self.stat:
            log_diff(_("File %s is missing"))
            return 0
        if self.type != other.type:
            log_diff(_("File %%s has type %s, expected %s") % (other.type, self.type))
            return 0

        if self.isreg() or self.isdir() or self.isfifo():
            if not self.perms_equal(other):
                log_diff(_("File %%s has permissions %s, expected %s") % (other.getperms(), self.getperms()))
                return 0
            if (int(self.stat.st_mtime) != int(other.stat.st_mtime)) and (
                self.stat.st_mtime > 0 or other.stat.st_mtime > 0
            ):
                log_diff(
                    _("File %%s has mtime %s, expected %s")
                    % (
                        dup_time.timetopretty(int(other.stat.st_mtime)),
                        dup_time.timetopretty(int(self.stat.st_mtime)),
                    )
                )
                return 0
            if self.isreg():
                if self.compare_data(other):
                    return 1
                else:
                    log_diff(_("Data for file %s is different"))
                    return 0
            else:
                return 1
        elif self.issym():
            if self.symtext == other.symtext or self.symtext + os.fsencode(os.sep) == other.symtext:
                return 1
            else:
                log_diff(_("Symlink %%s points to %s, expected %s") % (other.symtext, self.symtext))
                return 0
        elif self.isdev():
            if not self.perms_equal(other):
                log_diff(_("File %%s has permissions %s, expected %s") % (other.getperms(), self.getperms()))
                return 0
            if self.devnums != other.devnums:
                log_diff(_("Device file %%s has numbers %s, expected %s") % (other.devnums, self.devnums))
                return 0
            return 1
        assert 0, "Unhandled path type in ROPath.compare_verbose"

    def compare_data(self, other):
        """
        Compare data from two regular files, return true if same
        """
        f1 = self.open("rb")
        f2 = other.open("rb")

        def close():
            assert not f1.close(), "f1 failed to close"
            assert not f2.close(), "f2 failed to close"

        while True:
            buf1 = f1.read(config.copy_blocksize)
            buf2 = f2.read(config.copy_blocksize)
            if buf1 != buf2:
                close()
                return 0
            if not buf1:
                close()
                return 1

    def perms_equal(self, other):
        """
        True if self and other have same permissions and ownership
        """
        s1, s2 = self.stat, other.stat
        return self.mode == other.mode and s1.st_gid == s2.st_gid and s1.st_uid == s2.st_uid

    def copy(self, other):
        """
        Copy self to other.

        Also copies data. ``other`` must be a :class:`Path` instance.

        :param Path other: The destination path to copy to.
        """
        if self.isreg():
            other.writefileobj(self.open("rb"))
        elif self.isdir():
            os.mkdir(other.name)
        elif self.issym():
            os.symlink(self.symtext, other.name)
            if config.restore_ownership:
                os.lchown(other.name, self.stat.st_uid, self.stat.st_gid)
            other.setdata()
            return  # no need to copy symlink attributes
        elif self.isfifo():
            os.mkfifo(other.name)
        elif self.issock():
            socket.socket(socket.AF_UNIX).bind(other.name)
        elif self.isdev():
            if self.type == "chr":
                devtype = "c"
            else:
                devtype = "b"
            other.makedev(devtype, *self.devnums)
        self.copy_attribs(other)

    def copy_attribs(self, other):
        """
        Only copy attributes from self to other
        """
        if isinstance(other, Path):
            if not self.issym():
                if self.stat and config.restore_ownership:
                    util.maybe_ignore_errors(lambda: os.chown(other.name, self.stat.st_uid, self.stat.st_gid))
                util.maybe_ignore_errors(lambda: os.chmod(other.name, self.mode))
                util.maybe_ignore_errors(lambda: os.utime(other.name, (time.time(), self.stat.st_mtime)))
            other.setdata()
        else:
            # write results to fake stat object
            assert isinstance(other, ROPath), "'other' must be an ROPath instance"
            stat = StatResult()
            stat.st_uid, stat.st_gid = self.stat.st_uid, self.stat.st_gid
            stat.st_mtime = int(self.stat.st_mtime)
            other.stat = stat
            other.mode = self.mode

    def __str__(self):
        """
        Return string representation
        """
        return f"({util.uindex(self.index)} {self.type})"


class Path(ROPath):
    """
    Path class - wrapper around ordinary local files

    Besides caching stat() results, this class organizes various file
    code.
    """

    regex_chars_to_quote = re.compile('[\\\\\\"\\$`]')

    def rename_index(self, index):
        """
        Rename the given index if it matches a rename rule in config.

        :param tuple index: The index to rename.
        :returns: The renamed index or the original index if no rename found.
        :rtype: tuple
        """
        if not config.rename or not index:
            return index  # early exit
        path = os.path.normcase(os.path.join(*index))
        tail = []
        while path and path not in config.rename:
            path, extra = os.path.split(path)
            tail.insert(0, extra)
        if path:
            return config.rename[path].split(os.fsencode(os.sep)) + tail
        else:
            return index  # no rename found

    def __init__(self, base, index=()):
        """
        Path initializer
        """
        # self.opened should be true if the file has been opened, and
        # self.fileobj can override returned fileobj
        super().__init__(index)
        self.opened, self.fileobj = None, None
        if isinstance(base, str):
            base = os.fsencode(base)
        self.base = base

        # Create self.index, which is the path as a tuple
        self.index = self.rename_index(index)

        self.name = os.path.join(base, *self.index)

        # We converted any unicode base to filesystem encoding, so self.name should
        # be in filesystem encoding already and does not need to change
        self.uc_name = os.fsdecode(self.name)

        self.setdata()

    def setdata(self):
        """
        Refresh stat cache
        """
        try:
            # We may be asked to look at the target of symlinks rather than
            # the link itself.
            if config.copy_links:
                self.stat = os.stat(self.name)
            else:
                self.stat = os.lstat(self.name)
        except OSError as e:
            err_string = errno.errorcode[e.errno]
            if err_string in ["ENOENT", "ENOTDIR", "ELOOP", "ENOTCONN", "ENODEV"]:
                self.stat, self.type = None, None  # file doesn't exist
                self.mode = None
            else:
                raise
        else:
            self.set_from_stat()
            if self.issym():
                self.symtext = os.readlink(self.name)

    def append(self, ext):
        """
        Return new Path with ext added to index
        """
        if isinstance(ext, str):
            ext = os.fsencode(ext)
        return self.__class__(self.base, self.index + (ext,))

    def new_index(self, index):
        """
        Return new Path with index index
        """
        return self.__class__(self.base, index)

    def listdir(self):
        """
        Return list generated by os.listdir
        """
        return os.listdir(self.name)

    def isemptydir(self):
        """
        Return true if path is a directory and is empty
        """
        return self.isdir() and not self.listdir()

    def contains(self, child):
        """
        Return true if path is a directory and contains child
        """
        if isinstance(child, str):
            child = os.fsencode(child)
        # We don't use append(child).exists() here because that requires exec
        # permissions as well as read. listdir() just needs read permissions.
        return self.isdir() and child in self.listdir()

    def open(self, mode="rb"):
        """
        Return fileobj associated with self

        Usually this is just the file data on disk, but can be
        replaced with arbitrary data using the setfileobj method.
        """
        assert not self.opened, "Path is already marked as opened"
        if self.fileobj:
            result = self.fileobj
        else:
            result = open(self.name, mode)
        return result

    def makedev(self, type, major, minor):  # pylint: disable=redefined-builtin
        """
        Make a device file with specified type, major/minor nums
        """
        cmdlist = ["mknod", self.name, type, str(major), str(minor)]
        res = subprocess.run(cmdlist, check=False)
        if res.returncode != 0:
            raise PathException(f"Error running {cmdlist}")
        self.setdata()

    def mkdir(self):
        """
        Make directory(s) at specified path
        """
        log.Debug(_("Making directory %s") % self.uc_name)
        try:
            os.makedirs(self.name)
        except OSError:
            if not config.force:
                raise PathException(f"Error creating directory {self.uc_name}", 7)
        self.setdata()

    def delete(self):
        """
        Remove this file
        """
        log.Debug(_("Deleting %s") % self.uc_name)
        if self.isdir():
            util.ignore_missing(os.rmdir, self.name)
        else:
            util.ignore_missing(os.unlink, self.name)
        self.setdata()

    def touch(self):
        """
        Open the file, write 0 bytes, close
        """
        log.Debug(_("Touching %s") % self.uc_name)
        fp = self.open("wb")
        fp.close()

    def deltree(self):
        """
        Remove self by recursively deleting files under it
        """
        log.Debug(_("Deleting tree %s") % self.uc_name)
        shutil.rmtree(self.name)
        self.setdata()

    def get_parent_dir(self):
        """
        Return directory that self is in
        """
        if self.index:
            return Path(self.base, self.index[:-1])
        else:
            components = self.base.split(b"/")
            if len(components) == 2 and not components[0]:
                return Path(b"/")  # already in root directory
            else:
                return Path(b"/".join(components[:-1]))

    def writefileobj(self, fin):
        """
        Copy file object fin to self.  Close both when done.
        """
        fout = self.open("wb")
        while True:
            buf = fin.read(config.copy_blocksize)
            if not buf:
                break
            fout.write(buf)
        if fin.close() or fout.close():
            raise PathException("Error closing file object")
        self.setdata()

    def rename(self, new_path):
        """
        Rename the file at the current path to ``new_path``.

        :param Path new_path: The new path for the file.
        """
        shutil.move(self.name, new_path.name)
        self.setdata()
        new_path.setdata()

    def move(self, new_path):
        """
        Like rename but destination may be on different file system
        """
        self.copy(new_path)
        self.delete()

    def chmod(self, mode):
        """
        Change permissions of the path
        """
        os.chmod(self.name, mode)
        self.setdata()

    def get_temp_in_same_dir(self):
        """
        Return temp non existent path in same directory as self
        """
        global _tmp_path_counter
        parent_dir = self.get_parent_dir()
        while True:
            temp_path = parent_dir.append(f"duplicity_temp.{str(_tmp_path_counter)}")
            if not temp_path.type:
                return temp_path
            _tmp_path_counter += 1
            assert _tmp_path_counter < 10000, f"Warning too many temp files created for {self.uc_name}"

    def compare_recursive(self, other, verbose=None):
        """
        Compare self to other Path, descending down directories
        """
        from duplicity import selection  # TODO: avoid circ. dep. issue

        selfsel = selection.Select(self).set_iter()
        othersel = selection.Select(other).set_iter()
        return Iter.equal(selfsel, othersel, verbose)

    def __repr__(self):
        """
        Return string representation
        """
        return f"({self.index} {self.uc_name} {self.type})"

    def quote(self, s=None):
        """
        Return quoted version of s (defaults to self.name)

        The output is meant to be interpreted with shells, so can be
        used with subprocess.run(..., shell=True).
        """
        if not s:
            s = self.uc_name
        return '"%s"' % self.regex_chars_to_quote.sub(lambda m: f"\\{m.group(0)}", s)

    def unquote(self, s):
        """
        Return unquoted version of string s, as quoted by above quote()
        """
        assert s[0] == s[-1] == '"', "Input string must be enclosed in double quotes as produced by quote()"
        result = ""
        i = 1
        while i < len(s) - 1:
            if s[i] == "\\":
                result += s[i + 1]
                i += 2
            else:
                result += s[i]
                i += 1
        return result

    def get_filename(self):
        """
        Return filename of last component
        """
        components = self.name.split(b"/")
        assert components and components[-1], "Path has no final component (empty filename)"
        return components[-1]

    def get_canonical(self):
        """
        Return the canonical version of the path.

        Removes ".", and trailing slashes where possible.

        .. note::
           It's harder to remove "..", as "foo/bar/.." is not necessarily
           "foo", so we can't use ``os.path.normpath()``.

        :returns: The canonical path as a byte string.
        :rtype: bytes
        """
        newpath = b"/".join([x for x in self.name.split(b"/") if x and x != b"."])
        if self.uc_name[0] == "/":
            return b"/" + newpath
        elif newpath:
            return newpath
        else:
            return b"."


class DupPath(Path):
    """
    Represent duplicity data files

    Based on the file name, files that are compressed or encrypted
    will have different open() methods.
    """

    def __init__(self, base, index=(), parseresults=None):
        """
        Initialize a DupPath.

        The actual filename (no directory) must be the single element
        of the index, unless ``parseresults`` is given.

        :param str base: The base directory.
        :param tuple index: The path index.
        :param parseresults: Results of parsing the filename, if already done.
        """
        if parseresults:
            self.pr = parseresults
        else:
            assert len(index) == 1, "DupPath requires a single-element index when parseresults is not provided"
            self.pr = file_naming.parse(index[0])
            assert self.pr, "must be a recognizable duplicity file"

        Path.__init__(self, base, index)

    def filtered_open(self, mode="rb", gpg_profile=None):
        """
        Return a file object with appropriate encryption or compression.

        If encryption is specified but no ``gpg_profile`` is provided, use
        ``config.gpg_profile``.

        :param str mode: The mode to open the file in.
        :param gpg_profile: The GPG profile to use for encryption.
        :returns: A file-like object.
        :raises AssertionError: If the file is already open, or if it is marked
            as both encrypted and compressed.
        """
        assert (
            not self.opened and not self.fileobj
        ), "filtered_open called but file is already opened or has a custom fileobj"
        assert not (
            self.pr.encrypted and self.pr.compressed
        ), "A file cannot be both encrypted and compressed simultaneously"
        if gpg_profile:
            assert self.pr.encrypted, "gpg_profile provided but file is not marked as encrypted"

        if self.pr.compressed:
            return gzip.GzipFile(self.name, mode)
        elif self.pr.encrypted:
            if not gpg_profile:
                gpg_profile = config.gpg_profile
            if mode == "rb":
                return gpg.GPGFile(False, self, gpg_profile)
            elif mode == "wb":
                return gpg.GPGFile(True, self, gpg_profile)
        else:
            return self.open(mode)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
NcFTP backend.
"""

import os.path
import re
import urllib.error
import urllib.parse
import urllib.request

import duplicity.backend
from duplicity import config
from duplicity import errors
from duplicity import log
from duplicity import tempdir


class NCFTPBackend(duplicity.backend.Backend):
    """
    Connect to remote store using File Transfer Protocol
    """

    def __init__(self, parsed_url):
        """
        Initialize the NcFTP backend.

        :param ParsedUrl parsed_url: The parsed URL.
        :raises FatalError: If NcFTP is not found.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        # we expect an error return, so go low-level and ignore it
        try:
            ret, fout, ferr = self.subprocess_popen(["ncftpls", "-v"], ignore_errors=True)
        except Exception:
            fout = ""
            ret = -1
        # the expected error is 8
        if ret != 8 or not fout:
            raise errors.FatalError(
                "NcFTP not found:  Please install NcFTP version 3.1.9 or later",
                log.ErrorCode.ftp_ncftp_missing,
            )

        # version is the second word of the first line
        version = fout.split("\n")[0].split()[1]
        if version < "3.1.9":
            raise errors.FatalError(
                "NcFTP too old:  Duplicity requires NcFTP version 3.1.9,"
                "3.2.1 or later.  Version 3.2.0 will not work properly.",
                log.ErrorCode.ftp_ncftp_too_old,
            )
        elif version == "3.2.0":
            log.Warn(
                "NcFTP (ncftpput) version 3.2.0 may fail with duplicity.\n"
                "see: http://www.ncftpd.com/ncftp/doc/changelog.html\n"
                "If you have trouble, please upgrade to 3.2.1 or later",
                log.WarningCode.ftp_ncftp_v320,
            )
        log.Notice(f"NcFTP version is {version}")

        self.parsed_url = parsed_url

        self.url_string = duplicity.backend.strip_auth_from_url(self.parsed_url)

        # strip ncftp+ prefix
        self.url_string = duplicity.backend.strip_prefix(self.url_string, "ncftp")

        # This squelches the "file not found" result from ncftpls when
        # the ftp backend looks for a collection that does not exist.
        # version 3.2.2 has error code 5, 1280 is some legacy value
        self.popen_breaks["ncftpls"] = [5, 1280]

        # Use an explicit directory name.
        if self.url_string[-1] != "/":
            self.url_string += "/"

        self.password = self.get_password()

        if config.ftp_connection == "regular":
            self.conn_opt = "-E"
        else:
            self.conn_opt = "-F"

        self.tempfd, self.tempname = tempdir.default().mkstemp()
        self.tempfile = os.fdopen(self.tempfd, "w")
        self.tempfile.write(f"host {self.parsed_url.hostname}\n")
        self.tempfile.write(f"user {self.parsed_url.username}\n")
        self.tempfile.write(f"pass {self.password}\n")
        self.tempfile.close()
        self.flags = f"-f {self.tempname} {self.conn_opt} -t {config.timeout} -o useCLNT=0,useHELP_SITE=0 "
        if parsed_url.port is not None and parsed_url.port != 21:
            self.flags += f" -P '{parsed_url.port}'"

    def _put(self, source_path, remote_filename):
        remote_filename = os.fsdecode(remote_filename)
        remote_path = os.path.join(
            urllib.parse.unquote(re.sub("^/", "", self.parsed_url.path)),
            remote_filename,
        ).rstrip()
        commandline = f"ncftpput {self.flags} -m -V -C '{source_path.uc_name}' '{remote_path}'"
        self.subprocess_popen(commandline)

    def _get(self, remote_filename, local_path):
        remote_filename = os.fsdecode(remote_filename)
        remote_path = os.path.join(
            urllib.parse.unquote(re.sub("^/", "", self.parsed_url.path)),
            remote_filename,
        ).rstrip()
        commandline = (
            f"ncftpget {self.flags} -V -C '{self.parsed_url.hostname}' "
            f"'{remote_path.lstrip('/')}' '{local_path.uc_name}'"
        )
        self.subprocess_popen(commandline)

    def _list(self):
        """
        List files on the remote server using NcFTP.

        :returns: A list of filenames as byte strings.
        :rtype: list
        """
        # Do a long listing to avoid connection reset
        commandline = f"ncftpls {self.flags} -l '{self.url_string}'"
        _, l, _ = self.subprocess_popen(commandline)
        # Look for our files as the last element of a long list line
        return [os.fsencode(x.split()[-1]) for x in l.split("\n") if x and not x.startswith("total ")]

    def _delete(self, filename):
        """
        Delete a file from the remote server using NcFTP.

        :param str filename: The name of the file to delete.
        """
        commandline = f"ncftpls {self.flags} -l -X 'DELE {filename}' '{self.url_string}'"
        self.subprocess_popen(commandline)


duplicity.backend.register_backend("ncftp+ftp", NCFTPBackend)
duplicity.backend.uses_netloc.extend(["ncftp+ftp"])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
duplicity's gpg interface, builds upon python-gnupg
"""

import gzip
import locale
import os
import re
import sys
import tempfile
import threading

try:
    import gnupg
except ImportError:
    import sys
    import os

    print(f"DEBUG: sys.path={sys.path}", file=sys.stderr)
    print(f"DEBUG: sys.executable={sys.executable}", file=sys.stderr)
    print(f"DEBUG: os.environ['PYTHONPATH']={os.environ.get('PYTHONPATH')}", file=sys.stderr)
    raise

from duplicity import config
from duplicity import errors
from duplicity import log
from duplicity import tempdir
from duplicity import util

from hashlib import sha1
from hashlib import md5

blocksize = 256 * 1024


class GPGError(Exception):
    """
    Indicate some GPG Error
    """

    pass


class DuplicityGPG(gnupg.GPG):
    """
    Wrapper around gnupg.GPG to handle duplicity-specific logic.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize DuplicityGPG.
        """
        # python-gnupg 0.5.x calls --list-config during init, which gpgsm doesn't support.
        try:
            super().__init__(*args, **kwargs)
        except Exception:
            pass

    def make_args(self, args, passphrase):
        """
        Create the argument list for a GPG command.

        :param list args: Initial list of arguments.
        :param str passphrase: The GPG passphrase.
        :returns: The full list of arguments.
        :rtype: list
        """
        full_args = super().make_args(args, passphrase)
        if "_hidden_" in full_args:
            idx = full_args.index("_hidden_")
            if idx > 0 and full_args[idx - 1] == "--recipient":
                full_args.pop(idx)
                full_args.pop(idx - 1)
        if not passphrase:
            for i in range(len(full_args) - 1):
                if full_args[i] == "--pinentry-mode" and full_args[i + 1] == "loopback":
                    full_args.pop(i + 1)
                    full_args.pop(i)
                    break
        if config.use_gpgsm:
            for a in ["--fixed-list-mode", "--batch", "--with-colons"]:
                if a in full_args:
                    full_args.remove(a)
        return full_args

    def encrypt_file(self, fileobj_or_path, recipients, **kwargs):
        """
        Encrypt a file or file-like object.

        :param fileobj_or_path: The file to encrypt.
        :param list recipients: List of recipient key IDs.
        :param kwargs: Additional arguments for gnupg.GPG.encrypt_file.
        :returns: An encryption result object.
        """
        # python-gnupg 0.5.6 requires recipients or symmetric=True.
        # If we have hidden recipients in extra_args, we want to allow recipients to be empty.
        if not recipients and not kwargs.get("symmetric"):
            extra_args = kwargs.get("extra_args", [])
            if any("--hidden-recipient" in a for a in extra_args):
                recipients = ["_hidden_"]
        return super().encrypt_file(fileobj_or_path, recipients, **kwargs)


class GPGProfile(object):
    """
    Just hold some GPG settings, avoid passing tons of arguments
    """

    _version_re = re.compile(b"^gpg(sm)?( \\([^\\)]+\\))? (?P<maj>[0-9]+)\\.(?P<min>[0-9]+)\\.(?P<bug>[0-9]+)(-.+)?$")

    def __init__(self, passphrase=None, sign_key=None, recipients=None, hidden_recipients=None):
        """
        Set all data with initializer.

        ``passphrase`` is the passphrase. If it is None (not ""), assume
        it hasn't been set. ``sign_key`` can be blank if no signing is
        indicated, and ``recipients`` should be a list of keys. For all
        keys, the format should be a hex key like 'AA0E73D2'.

        :param str passphrase: The GPG passphrase.
        :param str sign_key: The hex ID of the signing key.
        :param list recipients: List of hex IDs for recipients.
        :param list hidden_recipients: List of hex IDs for hidden recipients.
        """
        assert passphrase is None or isinstance(
            passphrase, str
        ), f"passphrase must be a string or None, got {type(passphrase).__name__}"

        self.passphrase = passphrase
        self.signing_passphrase = passphrase
        self.sign_key = sign_key
        self.encrypt_secring = None
        if recipients is not None:
            assert isinstance(
                recipients, list
            ), f"recipients must be a list (not tuple), got {type(recipients).__name__}"  # must be list, not tuple
            self.recipients = recipients
        else:
            self.recipients = []

        if hidden_recipients is not None:
            assert isinstance(
                hidden_recipients, list
            ), f"hidden_recipients must be a list (not tuple), got {type(hidden_recipients).__name__}"
            self.hidden_recipients = hidden_recipients
        else:
            self.hidden_recipients = []

        self.gpg_version = self.get_gpg_version(config.gpg_binary)

    def get_gpg_version(self, binary):
        """
        Return the version of the GPG binary as a tuple.

        :param str binary: Path to the GPG binary.
        :returns: The version tuple (major, minor, bug).
        :rtype: tuple
        :raises GPGError: If the version cannot be determined.
        """
        binary = binary or "gpg"
        try:
            gpg = DuplicityGPG(gpgbinary=binary)
            if gpg.version:
                if isinstance(gpg.version, tuple):
                    return gpg.version
                return tuple(int(x) for x in re.split(r"[.-]", gpg.version)[:3])
        except Exception:
            pass

        # Fallback to manual check if python-gnupg fails (e.g. gpgsm)
        import subprocess

        try:
            args = [binary, "--version"]
            if config.gpg_options:
                args.extend(config.gpg_options.split())
            res = subprocess.run(args, capture_output=True)
            line = res.stdout.splitlines()[0]
            m = self._version_re.search(line)
            if m:
                return int(m.group("maj")), int(m.group("min")), int(m.group("bug"))
        except Exception:
            pass
        raise GPGError(f"failed to determine gnupg version of {binary}")


class GPGFile(object):
    """
    File-like object that encrypts decrypts another file on the fly
    """

    def __init__(self, encrypt, encrypt_path, profile):
        """
        GPGFile initializer

        If recipients is set, use public key encryption and encrypt to
        the given keys.  Otherwise, use symmetric encryption.

        encrypt_path is the Path of the gpg encrypted file.  Right now
        only symmetric encryption/decryption is supported.

        If passphrase is false, do not set passphrase - GPG program
        should prompt for it.
        """
        self.closed = False
        self.name = encrypt_path
        self.byte_count = 0
        self.signature = None
        self.encrypt = encrypt
        self.result = None
        self.thread_exception = None

        self.gpg = DuplicityGPG(gpgbinary=config.gpg_binary or "gpg")
        self.extra_args = ["--no-secmem-warning"]
        if not config.use_gpgsm:
            self.extra_args.append("--ignore-mdc-error")

        # user supplied options
        if config.gpg_options:
            for opt in config.gpg_options.split():
                self.extra_args.append(opt)

        passphrase = profile.passphrase or None
        if encrypt and profile.sign_key and profile.signing_passphrase:
            passphrase = profile.signing_passphrase

        # Support three versions of gpg present 1.x, 2.0.x, 2.1.x
        if profile.gpg_version[:1] == (1,):
            if config.use_agent:
                # gpg1 agent use is optional
                self.extra_args.append("--use-agent")
        elif profile.gpg_version[:2] >= (2, 1):
            if not config.use_agent and passphrase is not None and not config.use_gpgsm:
                # This forces gpg2 to ignore the agent.
                # Necessary to enforce truly non-interactive operation.
                self.extra_args.append("--pinentry-mode=loopback")

        self.gnupg_kwargs = {
            "passphrase": passphrase,
        }
        if encrypt:
            self.gnupg_kwargs["armor"] = False

        r, w = os.pipe()
        if encrypt:
            if profile.sign_key:
                self.gnupg_kwargs["sign"] = profile.sign_key

            if profile.recipients or profile.hidden_recipients:
                self.gnupg_kwargs["recipients"] = profile.recipients
                for h in profile.hidden_recipients:
                    self.extra_args.extend(["--hidden-recipient", h])
            else:
                self.gnupg_kwargs["symmetric"] = True

            if not config.use_gpgsm:
                # use integrity protection
                self.extra_args.append("--force-mdc")

            # Turn off compression if needed
            if not config.compression:
                self.extra_args.append("--compress-algo=none")

            self.gpg_input = os.fdopen(w, "wb")
            self.pipe_read = os.fdopen(r, "rb")
            self.output_file = encrypt_path.open("wb")

            def on_data(data):
                self.output_file.write(data)
                return False

            self.gpg.on_data = on_data

            def run_gpg():
                try:
                    recipients = self.gnupg_kwargs.pop("recipients", None)
                    self.result = self.gpg.encrypt_file(
                        self.pipe_read, recipients, extra_args=self.extra_args, **self.gnupg_kwargs
                    )
                except Exception as e:
                    self.thread_exception = e
                    log.Debug(f"GPG encryption thread failed: {str(e)}")
                finally:
                    self.pipe_read.close()
                    self.output_file.close()

            self.thread = threading.Thread(target=run_gpg)
            self.thread.start()
        else:
            if (profile.recipients or profile.hidden_recipients) and profile.encrypt_secring:
                self.extra_args.extend(["--secret-keyring", profile.encrypt_secring])

            self.gpg_output = os.fdopen(r, "rb")
            self.pipe_write = os.fdopen(w, "wb")
            self.input_file = encrypt_path.open("rb")

            def on_data(data):
                if data:
                    self.pipe_write.write(data)
                else:
                    try:
                        self.pipe_write.close()
                    except (IOError, ValueError):
                        pass
                return False

            self.gpg.on_data = on_data

            def run_gpg():
                try:
                    self.result = self.gpg.decrypt_file(
                        self.input_file, extra_args=self.extra_args, **self.gnupg_kwargs
                    )
                except Exception as e:
                    self.thread_exception = e
                    log.Debug(f"GPG decryption thread failed: {str(e)}")
                finally:
                    self.pipe_write.close()
                    self.input_file.close()

            self.thread = threading.Thread(target=run_gpg)
            self.thread.start()

    def read(self, length=-1):
        """
        Read decrypted data from GPG.

        :param int length: Number of bytes to read.
        :returns: The data read.
        :rtype: bytes
        """
        try:
            res = self.gpg_output.read(length)
            if res is not None:
                self.byte_count += len(res)
        except Exception:
            self.gpg_failed()
        return res

    def write(self, buf):
        """
        Write data to be encrypted by GPG.

        :param bytes buf: Data to write.
        :returns: Number of bytes written.
        :rtype: int
        """
        try:
            res = self.gpg_input.write(buf)
            if res is not None:
                self.byte_count += len(buf)
        except Exception:
            self.gpg_failed()
        return res

    def tell(self):
        """
        Return the current byte count processed.

        :returns: Current byte count.
        :rtype: int
        """
        return self.byte_count

    def seek(self, offset):
        """
        Seek to a given offset (only forward seeking during decryption).

        :param int offset: Offset to seek to.
        """
        assert not self.encrypt, "seek() is only supported when decrypting (encrypt=False)"
        assert offset >= self.byte_count, f"{int(offset)} < {int(self.byte_count)}"
        if offset > self.byte_count:
            self.read(offset - self.byte_count)

    def gpg_failed(self):
        """
        Handle a GPG failure by collecting error information and raising GPGError.

        :raises GPGError: Always.
        """
        if self.thread.is_alive():
            self.thread.join()
        msg = "GPG Failed, see log below:\n"
        msg += "===== Begin GnuPG log =====\n"
        if self.result is not None:
            msg += f"{self.result.stderr}\n"
        if self.thread_exception is not None:
            msg += f"Thread exception: {str(self.thread_exception)}\n"
        msg += "===== End GnuPG log =====\n"
        if not (msg.find("invalid packet (ctb=14)") > -1):
            raise GPGError(msg)
        else:
            return ""

    def close(self):
        """
        Close the GPGFile and wait for the GPG process to finish.
        """
        if self.encrypt:
            try:
                self.gpg_input.close()
            except Exception:
                self.gpg_failed()
        else:
            try:
                while True:
                    res = self.gpg_output.read(blocksize)
                    if not res:
                        break
                self.gpg_output.close()
            except Exception:
                self.gpg_failed()

        self.thread.join()
        if self.result is None or not self.result:
            if config.use_gpgsm and self.result is not None and self.result.returncode == 0:
                pass
            else:
                self.gpg_failed()

        if not self.encrypt:
            self.set_signature()
        self.closed = True

    def set_signature(self):
        """
        Set self.signature to signature keyID

        This only applies to decrypted files.  If the file was not
        signed, set self.signature to None.
        """
        if self.result and self.result.key_id and (self.result.fingerprint or self.result.valid):
            self.signature = self.result.key_id
        else:
            self.signature = None

    def get_signature(self):
        """
        Return  keyID of signature, or None if none
        """
        assert self.closed, "file not closed yet"
        return self.signature


def GPGWriteFile(block_iter, filename, profile, size=200 * 1024 * 1024, max_footer_size=16 * 1024):
    """
    Write GPG compressed file of given size

    This function writes a gpg compressed file by reading from the
    input iter and writing to filename.  When it has read an amount
    close to the size limit, it "tops off" the incoming data with
    incompressible data, to try to hit the limit exactly.

    block_iter should have methods .next(size), which returns the next
    block of data, which should be at most size bytes long.  Also
    .get_footer() returns a string to write at the end of the input
    file.  The footer should have max length max_footer_size.

    Because gpg uses compression, we don't assume that putting
    bytes_in bytes into gpg will result in bytes_out = bytes_in out.
    However, do assume that bytes_out <= bytes_in approximately.

    Returns true if succeeded in writing until end of block_iter.
    """

    # workaround for circular module imports
    from duplicity import path

    def top_off(bytelen, file):
        """
        Add bytelen of incompressible data to to_gpg_fp

        In this case we take the incompressible data from the
        beginning of filename (it should contain enough because size
        >> largest block size).
        """
        if bytelen <= 0:
            return
        incompressible_fp = open(filename, "rb")
        try:
            while bytelen > 0:
                incompressible_fp.seek(0)
                written = util.copyfileobj(incompressible_fp, file.gpg_input, bytelen)
                if written > 0:
                    bytelen -= written
                else:
                    # Fallback if file is empty
                    import os

                    data = os.urandom(min(bytelen, 64 * 1024))
                    file.gpg_input.write(data)
                    bytelen -= len(data)
        finally:
            incompressible_fp.close()

    def get_current_size():
        return os.stat(filename).st_size

    target_size = size - 300 * 1024  # fudge factor, compensate for gpg buffering
    data_size = target_size - max_footer_size
    file = GPGFile(True, path.Path(filename), profile)
    at_end_of_blockiter = 0
    try:
        while True:
            read_size = block_iter.get_read_size()
            cur_size = get_current_size()
            bytes_to_go = data_size - cur_size
            if bytes_to_go < read_size:
                break
            try:
                data = block_iter.__next__().data
            except StopIteration:
                at_end_of_blockiter = 1
                break
            except Exception as e:
                raise errors.FatalError(f"Read error on {os.fsdecode(filename)}: {str(e)}")
            file.write(data)

        file.write(block_iter.get_footer())
        if not at_end_of_blockiter:
            # don't pad last volume
            cursize = get_current_size()
            if cursize < target_size:
                top_off(target_size - cursize, file)
        file.close()
        return at_end_of_blockiter
    except Exception:
        # ensure that GPG processing terminates
        file.close()
        raise


def GzipWriteFile(block_iter, filename, size=200 * 1024 * 1024, gzipped=True):
    """
    Write a gzipped compressed file of the given size.

    This is like :func:`GPGWriteFile` except it writes a gzipped
    file instead of a GPG encrypted file.

    The input requirements on ``block_iter`` and the output is the same as
    :func:`GPGWriteFile` (returns true if wrote until end of ``block_iter``).

    :param block_iter: Iterator yielding data blocks.
    :param str filename: The name of the file to write.
    :param int size: The target size of the file.
    :param bool gzipped: Whether to compress the file with gzip.
    :returns: True if the end of the block iterator was reached.
    :rtype: bool
    """

    class FileCounted(object):
        """
        Wrapper around file object that counts number of bytes written
        """

        def __init__(self, fileobj):
            self.fileobj = fileobj
            self.byte_count = 0

        def write(self, buf):
            result = self.fileobj.write(buf)
            self.byte_count += len(buf)
            return result

        def flush(self):
            return self.fileobj.flush()

        def close(self):
            return self.fileobj.close()

    target_size = size - 50 * 1024
    file_counted = FileCounted(open(filename, "wb"))

    # if gzipped wrap with GzipFile else plain file out
    if gzipped:
        outfile = gzip.GzipFile(None, "wb", 6, file_counted)
        size -= 32 * 1024
    else:
        outfile = file_counted
    at_end_of_blockiter = 0
    while True:
        bytes_to_go = target_size - file_counted.byte_count
        if bytes_to_go < block_iter.get_read_size():
            break
        try:
            new_block = next(block_iter)
        except StopIteration:
            at_end_of_blockiter = 1
            break
        outfile.write(new_block.data)
        outfile.flush()

    outfile.write(block_iter.get_footer())
    assert not outfile.close(), "outfile failed to close"
    assert not file_counted.close(), "file_counted failed to close"
    return at_end_of_blockiter


def PlainWriteFile(block_iter, filename, size=200 * 1024 * 1024, gzipped=False):
    """
    Write a plain uncompressed file of the given size.

    This is like :func:`GPGWriteFile` except it writes a plain
    file instead of a GPG encrypted file.

    :param block_iter: Iterator yielding data blocks.
    :param str filename: The name of the file to write.
    :param int size: The target size of the file.
    :param bool gzipped: Should be False for plain files.
    :returns: True if the end of the block iterator was reached.
    :rtype: bool
    """
    return GzipWriteFile(block_iter, filename, size, gzipped)


def get_hash(hash, path, hex=True):  # pylint: disable=redefined-builtin
    """
    Return the hash of the file at the given path.

    :param str hash: The hash algorithm to use ("MD5" or "SHA1").
    :param Path path: The path to the file.
    :param bool hex: If True, return the hash in hexadecimal form.
    :returns: The calculated hash.
    :rtype: str or bytes
    """
    # assert path.isreg()
    fp = path.open("rb")
    if hash == "SHA1":
        hash_obj = sha1(usedforsecurity=False)
    elif hash == "MD5":
        hash_obj = md5(usedforsecurity=False)
    else:
        assert 0, f"Unknown hash {hash}"

    while True:
        buf = fp.read(blocksize)
        if not buf:
            break
        hash_obj.update(buf)
    assert not fp.close(), "fp failed to close"
    if hex:
        return hash_obj.hexdigest()
    else:
        return hash_obj.digest()
